from time import time, sleep
from enum import Enum
from xml.etree.ElementTree import PI

from PyQt5.QtSerialPort import QSerialPort, QSerialPortInfo
from PyQt5.QtCore import QEventLoop, QObject, QTimer, QThread, pyqtSignal

from . import gc_constants, GSI_Parser

class CommunicationConstants:
    GSI = "GSI"
    GEOCOM = "geoCOM"

    GSI_MESSAGE_PREFIX = "?"
    GEOCOM_MESSAGE_PREFIX = "%R1Q"

    GEOCOM_REPLY_PREFIX = "%R1P"

    GSI_REPLY_PREFIXES = ["?", "*"]

    class ImplementationStates(Enum):
        NOT_YET_DETECTED = 0
        NOT_IMPLEMENTED = 1
        GSI = 2
        GEOCOM = 3

    COMMANDS = [
        ""
    ]


class TachyCommand(QObject):
    MESSAGE_PREFIX = ""
    REPLY_PREFIX = ""
    protocol = None

    def __init__(self, command: str, label = None, time_out = 2, args = []) -> None:
        super().__init__()
        self.command = command
        self.args = args
        self.transaction_id = 0
        self.time_out = time_out
        self.label = command if label is None else label


    def parse_reply(self, reply_bytes: bytes) -> str:
        return reply_bytes.decode('ascii')

    def set_transaction_id(self, id: int) -> None:
        self.transaction_id = id

    def get_transaction_id(self) -> int:
        return self.transaction_id

    def get_protocol(self):
        return self.protocol

    def __str__(self):
        try:
            return f"{self.protocol}: {self.command}, {str(self.args)}"
        except ValueError:
            return "Unitialized command object"


class GSICommand(TachyCommand):
    def __init__(self, command: str, label = None, time_out = 2, args = []) -> None:
        super().__init__(command, label, time_out=time_out, args=args)
        self.protocol=CommunicationConstants.GSI

    def get_transaction_id(self) -> int:
        return 1

    @property
    def bytes(self):
        params = ""
        if len(self.args):
            for arg in self.args:
                params += (str(arg) + " ")

        message = f"{self.command}{params}{gc_constants.CRLF}"
        return message.encode('ascii')


class GeoCOMCommand(TachyCommand):
    MESSAGE_PREFIX = CommunicationConstants.GEOCOM_MESSAGE_PREFIX

    def __init__(self, command: str, label = None, time_out = 2, *args) -> None:
        super().__init__(command, label, time_out=time_out, args = args)
        self.protocol=CommunicationConstants.GEOCOM

    @property
    def bytes(self):
        message = f"{self.MESSAGE_PREFIX},{str(self.command)},{self.transaction_id}:{','.join(map(str, self.args))}{gc_constants.CRLF}"
        return message.encode('ascii')


class TachyReply:
    def __init__(self, bites: bytes) -> None:
        self.bites = bites
        self.ascii = bites.data().decode('ascii')

    def __str__(self) -> str:
        return self.ascii

    def get_transaction_id(self) -> int:
        raise ValueError("This base type is not supposed to have an id.")

    def get_protocol(self) -> str:
        if self.ascii[0] in CommunicationConstants.GSI_REPLY_PREFIXES:
            return CommunicationConstants.GSI
        if self.ascii.startswith(CommunicationConstants.GEOCOM_REPLY_PREFIX) or self.ascii[:4] == "@W127":
            return CommunicationConstants.GEOCOM
        raise ValueError(f"Tachymeter reply uses unknown protocoll: {self.ascii}")

    def get_result(self):
        raise NotImplemented("Has to be implemented for each subclass.")

    def success(self):
        return False


class GSIReply(TachyReply):
    @property
    def PREFIX(self):
        self.ascii[0]

    def get_transaction_id(self) -> int:
        return 1

    def success(self):
        return self.ascii.startswith(GSI_Parser.REPLY_ACK)

    def get_result(self):
        return GSI_Parser.parse(self.ascii)


class GeoCOMReply(TachyReply):
    PREFIX = CommunicationConstants.GEOCOM_REPLY_PREFIX

    def get_transaction_id(self) -> int:
        header = self.ascii.split(':')[0]
        segments = header.split(",")
        return int(segments[2])

    def success(self):
        return self.ascii.startswith(GeoCOMReply.PREFIX)

    def get_result(self):
        payload = self.ascii.split(':')[1].strip()
        return payload.split(",")


class MessageQueue(QObject):
    non_requested_data = pyqtSignal(str)
    def __init__(self, n_slots=7):
        super().__init__()
        self.indices = list(range(1, n_slots + 1))
        self.slots = {}
        self.serial = QSerialPort

    def set_serial(self, serial):
        self.serial = serial

    def append(self, msg: TachyCommand) -> int:
        def first_free_slot() -> int:
            for i in self.indices:
                if i not in self.slots.keys():
                    return i
            return False
        if self.serial is None:
            return False
        slot = first_free_slot()
        if slot and self.serial:
            self.slots[slot] = {"message": msg.label,
                                "timeout": time() + msg.time_out}
            msg.set_transaction_id(slot)
            self.serial.write(msg.bytes)
        return slot

    def check_timeouts(self):
        over_ripes = list(filter(lambda i: i[1]['timeout'] < time(), self.slots.items()))
        messages = []
        for index, msg in over_ripes:
            messages.append((index, self.slots.pop(index)['message']))
        return messages

    def register_reply(self, reply: TachyReply) -> tuple:
        message_id = reply.get_transaction_id()
        try:
            request = self.slots.pop(message_id)
            return request, reply
        except KeyError:
            self.non_requested_data.emit(reply.ascii)

    def close(self):
        self.serial.close()

    def __str__(self) -> str:
        n_slots = self.indices[-1]
        return f"Message queue with {n_slots} slot{'s' if n_slots > 1 else ''}"


class Dispatcher(QThread):
    pollingInterval = 1000
    serial_disconnected = pyqtSignal(str)
    serial_connected = pyqtSignal()
    timed_out = pyqtSignal(str)
    log = pyqtSignal(str)
    non_requested_data = pyqtSignal(str)

    def __init__(self, gsi_queue: MessageQueue, geocom_queue: MessageQueue, reply_handler, parent = None) -> None:
        super(self.__class__, self).__init__(parent)
        self.pollingTimer = QTimer(self)
        self.pollingTimer.timeout.connect(self.poll)
        self.serial = QSerialPort()
        self.queues = {CommunicationConstants.GSI: gsi_queue,
                       CommunicationConstants.GEOCOM: geocom_queue}
        self.reply_types = {CommunicationConstants.GSI: GSIReply,
                            CommunicationConstants.GEOCOM: GeoCOMReply}
        self.reply_handler = reply_handler
        for queue in self.queues.values():
            queue.non_requested_data.connect(self.emit_non_requested_data)

    def emit_non_requested_data(self, data):
        self.non_requested_data.emit(data)

    def hook_up(self):
        # close port if connection is established
        self.log.emit("Connection attempt")
        if self.serial.isOpen():
            port_name = self.serial.portName
            self.serial.close()
            self.serial_disconnected.emit(port_name)
            return
        port_names = [port.portName() for port in QSerialPortInfo.availablePorts()]
        port_names.reverse()
        def moin():
            if port_names:
                ping = Ping(port_names.pop(), 500)
                ping.timed_out.connect(moin)
                ping.found_tachy.connect(self.set_serial_port)
                ping.fire()
        moin()

    def start(self):
        self.pollingTimer.start(self.pollingInterval)
        self.serial_connected.emit()
        self.log.emit("Started listening.")
        loop = QEventLoop()
        loop.exec_()
        
    def stop(self):
        self.pollingTimer.stop()

    def send_log(self, *args):
        self.log.emit(str(args))

    def set_serial_port(self, port_name):
        if self.serial.isOpen():
            self.serial.close()
        self.serial.setPortName(port_name)
        self.serial.open(QSerialPort.ReadWrite)
        for queue in self.queues.values():
            self.log.emit(f"Connecting {str(queue)} to {self.serial.portName()}")
            queue.set_serial(self.serial)
        self.start()

    def poll(self):
        if self.serial.error() == QSerialPort.ResourceError:  # device is unexpectedly removed from the system
            self.serial.close()
            self.serial_disconnected.emit(f"Connection {self.serial.portName} failed")
        if self.serial.canReadLine():
            reply = TachyReply(self.serial.readLine())
            self.register_reply(reply)
        else:
            pass
        timed_out = [q.check_timeouts() for q in self.queues.values()]
        if any(timed_out):
            description = f"Timed out: {str(list(timed_out))}"
            self.timed_out.emit(description)

    def send(self, message: TachyCommand):
        queue = self.queues[message.protocol]
        success = queue.append(message)
        return bool(success)

    def register_reply(self, reply: TachyReply):
        protocol = reply.get_protocol()
        reply_type = self.reply_types[protocol]
        queue = self.queues[protocol]
        try:
            self.reply_handler.handle(*queue.register_reply(reply_type(reply.bites)))
        except TypeError:
            """TypeErrors usually stem from incoming data that can not be
            associated with any request. These are handled by the 'non_requested_data'
            signal."""
            pass

class Ping(QThread):
    found_tachy = pyqtSignal(str)
    found_something = pyqtSignal(str)
    timed_out = pyqtSignal(str)
    pinging = pyqtSignal(str)

    def __init__(self, port_name, timeout=1200):
        super().__init__()
        self.timeout = timeout
        self.serial = QSerialPort()
        self.serial.setPortName(port_name)
        self.serial.open(QSerialPort.ReadWrite)
        message = f"{GeoCOMCommand.MESSAGE_PREFIX},{str(gc_constants.BMM_BeepAlarm)},1:{gc_constants.CRLF}"
        self.serial.writeData(message.encode('ascii'))
        self.timer = QTimer(self)
        self.timer.setSingleShot(True)
        self.timer.timeout.connect(self.read)

    def fire(self):
        self.timer.start(self.timeout)
        loop = QEventLoop()
        loop.exec_()

    def read(self):
        self.pinging.emit(f"Pinging {self.serial.portName()}")
        if self.serial.canReadLine():
            reply = bytes(self.serial.readLine()).decode('ascii')
            self.serial.close()
            if reply.startswith(GeoCOMReply.PREFIX):
                self.found_tachy.emit(self.serial.portName())
            else:
                self.found_something.emit(self.serial.portName())
        else:
            self.timed_out.emit(f"Ping timed out: {self.serial.portName()}")

        self.serial.close()
        self.quit()




from socket import timeout
from PyQt5.QtCore import pyqtSignal, QObject, QEventLoop

from tachyconnect.ts_control import TachyReply


class ReplyHandler(QObject):
    caught_reply = pyqtSignal(TachyReply)
    fall_back_signal = pyqtSignal(tuple)
    has_fall_back = False
    slots = {}

    def __init__(self, fall_back = None):
        super().__init__()
        if fall_back:
            self.fall_back_signal.connect(fall_back)
            self.has_fall_back = True

    def register_command(self, command_class, slot):
        self.slots[command_class.__name__] = slot

    def unregister_command(self, command_class):
        self.slots.pop(command_class.__name__, None)

    def handle(self, request, reply):
        print(request)
        print(reply)
        self.caught_reply.emit(reply)
        command = request['message']
        slot = self.slots.get(command)
        if slot:
            slot(*reply.get_result())
            return True
        elif self.has_fall_back:
            self.fall_back_signal.emit((request, reply))
            return True
        return False

class CommandChain:
    def __init__(self, dispatcher, *commands):
        self.dispatcher = dispatcher
        self.reply_handler = self.dispatcher.reply_handler
        self.index = 0
        self.loop = QEventLoop()
        self.commands = commands

    def set_commands(self, *args):
        print('setting commands: ' + args)
        self.commands = args

    def run_chain(self, *results):
        command = self.commands[self.index]
        if results:
            self.reply_handler.unregister_command(self.commands.command)
            self.index += 1
        if self.index < len(self.commands):
            command = self.commands[self.index]
            self.reply_handler.register_command(command.command, self.run_chain)
            self.dispatcher.send(command.command(timeout=command.timeout, args=command.args).get_geocom_command())

    def reset(self):
        self.index = 0

class ChainableCommand:
    def __init__(self, command, timeout=2, *args):
        self.command = command
        self.timeout = timeout
        self.args = args
        print(self.command)
        print(self.timeout)
        print(self.args)
        print('ChainableCommand ctored.')





from . import gc_constants as gc
from PyQt5.QtCore import QObject, pyqtSignal, pyqtSlot, QTimer, QThread
from PyQt5.QtSerialPort import QSerialPort
from time import time


class GeoCOMRequest:
    PREFIX = "%R1Q"
    TERMINATOR = "\r\n"

    def __init__(self, command, *args):
        self.command = str(command)
        self.args = ','.join(map(str, args))
        self.transaction_id = 0

    def set_transaction_id(self, slot):
        self.transaction_id = slot

    def __str__(self):
        msg = ",".join((GeoCOMRequest.PREFIX, self.command))
        if self.transaction_id:
            msg = ",".join((msg, str(self.transaction_id)))
        msg = ":".join((msg, self.args))
        msg += GeoCOMRequest.TERMINATOR
        return msg

    @property
    def bytes(self):
        return str(self).encode("ascii")

    __repr__ = __str__


class GeoCOMReply:
    PREFIX = "%R1P"

    def __init__(self, bites):
        self.msg = bites.decode('ascii')
        head, tail = self.msg.split(':')
        head = head.split(',')
        tail = tail.strip().split(',')
        self.com_code = int(head[1])
        self.transaction_id = int(head[2])
        self.ret_code = int(tail.pop(0))
        self.results = tail

    def __str__(self):
        return self.msg

    __repr__ = __str__



class GeoCOMReplyHandler:
    def __init__(self, command, types, signal):
        self.command = command
        self.types = types
        self.signal = signal

    def handle(self, reply):
        # Will not work because of type mismatch between reply retcodes
        # (str) and lib constants (int) but looks nicer this way.
        if reply.com_code == gc.GRC_OK and reply.ret_code == gc.GRC_OK:
            casted = [t(value) for t, value in zip(self.types, reply.results)]
            self.signal.emit(casted)


class GeoCOMCallCenter:
    """This is the skeleton of a prototype that mainly exists to describe
    a design. Use at your own risk.
    """
    def __init__(self):
        self.handlers = {}

    def register(self, handler):
        self.handlers[str(handler.command)] = handler

    @pyqtSlot(GeoCOMRequest, GeoCOMReply)
    def handle(self, request, reply):
        handler = self.handlers.get(request.command, False)
        if handler:
            handler.handle(reply)


@pyqtSlot(str)
def connect_beep(port_name):
    serial = QSerialPort()
    serial.setPortName(port_name)
    serial.open(QSerialPort.WriteOnly)
    message = "%R1Q," + str(gc.BMM_BeepAlarm) + ":" + gc.CRLF
    serial.writeData(message.encode('ascii'))
    serial.close()



# -*- coding: utf-8 -*-
# This parser uses the dictionararies defined by 
#
from PyQt5.QtCore import QThread, pyqtSignal, QTimer
from PyQt5.QtSerialPort import QSerialPort

dict_projections = {
    "WGS84": 4326,
    "Gauss-Krüger1": 5520,
    "Gauss-Krüger2": 31466,
    "Gauss-Krüger3": 31467,
    "Gauss-Krüger4": 31468,
    "Gauss-Krüger5": 31469,
    "ETRS32 8stellig": 102329,
    "ETRS32 7stellig": 102328,
    "ETRS32 6stellig": 25832,
    "ETRS33 8stellig": 102359,
    "ETRS33 7stellig": 102360,
    "ETRS33 6stellig": 25833,
    "Kein Projektionssystem": "NONE",
    }

dict_labels = {
# Textinormationen
    "@T": "manual_notes",
    "01": "manual_notes",
    "02": "not_defined",
    "03": "not_defined",
    "04": "not_defined",
    "05": "not_defined",
    "06": "not_defined",
    "07": "not_defined",
    "08": "not_defined",
    "09": "not_defined",
    "10": "not_defined",
# Allgemeines
    "11": "pointId",
    "12": "serialID",
    "13": "identInstrument",
    "14": "not_defined",
    "15": "not_defined",
    "16": "stationID",
    "17": "date",
    "18": "time1",
    "19": "time2",
# Richtungen
    "20": "angularAccuracy",
    "21": "hzAngle",
    "22": "vertAngle",
    "23": "not_defined",
    "24": "targetAngle",
    "25": "AngleDifference",
    "26": "rotationAngle",
    "27": "targetVertAngle",
    "28": "vertDifference",
    "29": "not_defined",
# Distanzen
    "30": "not_defined",
    "31": "slopeDist",
    "32": "hzDist",
    "33": "HeightDiff",
    "34": "targetDist",
    "35": "DistDiffHz",
    "36": "targetDiffHeight",
    "37": "HeightDiff",
    "38": "targetSlopeDist",
    "39": "DistDiffSlope",
# Codeblock
    "40": "not_defined",
    "41": "CodeBlockID",
    "42": "stationID",
    "43": "stationHeight",
    "44": "Info3",
    "45": "Info4",
    "46": "Info5",
    "47": "Info6",
    "48": "Info7",
    "49": "Info8",
# Distanz-Zusatzinformationen
    "50": "not_defined",
    "51": "constDist",
    "52": "constPrism",
    "53": "MeasureNr",
    "54": "Signalqual",
    "55": "refractCoef",
    "56": "relHumidity",
    "57": "SigDist",
    "58": "PrismConstant",
    "59": "PPM",
# Richtungs-Zusatzinformationen
    "60": "not_defined",
    "61": "longInclin",
    "62": "transInclin",
    "63": "standAxisTilt",
    "64": "standAxisAz",
    "65": "not_defined",
    "66": "not_defined",
    "67": "not_defined",
    "68": "not_defined",
    "69": "not_defined",
# Bemerkungen
    "70": "not_defined",
    "71": "pointCode",
    "72": "note2",
    "73": "note3",
    "74": "note4",
    "75": "note5",
    "76": "note6",
    "77": "note7",
    "78": "note8",
    "79": "note9",
# Koordinaten
    "80": "not_defined",
    "81": "targetX",
    "82": "targetY",
    "83": "targetZ",
    "84": "stationY",
    "85": "stationX",
    "86": "stationZ",
    "87": "reflectorZ",
    "88": "instrumentZ",
    "89": "not_defined",
# Rohdaten
    "90": "not_defined",
    "91": "DirectionUncorr",
    "92": "VerticalUncorr",
    "93": "inclinationLenght",
    "94": "inclinationCross",
    "95": "not_defined",
    "96": "not_defined",
    "97": "not_defined",
    "98": "not_defined",
    "99": "not_defined",
    }

dict_formats = {
    "@T": "TEXT",
    "01": "TEXT",
    "02": "TEXT",
    "03": "TEXT",
    "04": "TEXT",
    "05": "TEXT",
    "06": "TEXT",
    "07": "TEXT",
    "08": "TEXT",
    "09": "TEXT",
    "10": "TEXT",
    "11": "TEXT",
    "12": "TEXT",
    "13": "TEXT",
    "14": "TEXT",
    "15": "TEXT",
    "16": "TEXT",
    "40": "TEXT",
    "41": "TEXT",
    "42": "TEXT",
    "43": "TEXT",
    "44": "TEXT",
    "45": "TEXT",
    "46": "TEXT",
    "47": "TEXT",
    "48": "TEXT",
    "49": "TEXT",
    "70": "TEXT",
    "71": "TEXT",
    "72": "TEXT",
    "73": "TEXT",
    "74": "TEXT",
    "75": "TEXT",
    "76": "TEXT",
    "77": "TEXT",
    "78": "TEXT",
    "79": "TEXT",
    "17": "DATE",
    "18": "DATE",
    "19": "DATE",
    "20": "DOUBLE",
    "21": "DOUBLE",
    "22": "DOUBLE",
    "23": "DOUBLE",
    "24": "DOUBLE",
    "25": "DOUBLE",
    "26": "DOUBLE",
    "27": "DOUBLE",
    "28": "DOUBLE",
    "29": "DOUBLE",
    "30": "DOUBLE",
    "31": "DOUBLE",
    "32": "DOUBLE",
    "33": "DOUBLE",
    "34": "DOUBLE",
    "35": "DOUBLE",
    "36": "DOUBLE",
    "37": "DOUBLE",
    "38": "DOUBLE",
    "39": "DOUBLE",
    "50": "DOUBLE",
    "51": "DOUBLE",
    "52": "DOUBLE",
    "53": "DOUBLE",
    "54": "DOUBLE",
    "55": "DOUBLE",
    "56": "DOUBLE",
    "57": "DOUBLE",
    "58": "DOUBLE",
    "59": "DOUBLE",
    "60": "DOUBLE",
    "61": "DOUBLE",
    "62": "DOUBLE",
    "63": "DOUBLE",
    "64": "DOUBLE",
    "65": "DOUBLE",
    "66": "DOUBLE",
    "67": "DOUBLE",
    "68": "DOUBLE",
    "69": "DOUBLE",
    "80": "DOUBLE",
    "81": "DOUBLE",
    "82": "DOUBLE",
    "83": "DOUBLE",
    "84": "DOUBLE",
    "85": "DOUBLE",
    "86": "DOUBLE",
    "87": "DOUBLE",
    "88": "DOUBLE",
    "89": "DOUBLE",
    "90": "DOUBLE",
    "91": "DOUBLE",
    "92": "DOUBLE",
    "93": "DOUBLE",
    "94": "DOUBLE",
    "95": "DOUBLE",
    "96": "DOUBLE",
    "97": "DOUBLE",
    "98": "DOUBLE",
    "99": "DOUBLE",
    }


dict_units_attributes_digits = {
    "0": "meter_1mm",
    "1": "feet",
    "2": "gon",
    "3": "degree_decimal",
    "4": "degree_sexagesimal",
    "5": "mil",
    "6": "meter_1_10mm",
    "7": "feet_1_10t",
    "8": "meter_1_100mm",
    ".": None
    }

# number of decimal places for each type of unit
dict_units_digits = {
    "0": 3,
     "1": 3,
     "2": 5,
     "3": 5,
     "4": 5,
     "5": 4,
     "6": 4,
     "7": 4,
     "8": 5,
     "9": 0,
     ".": 0,
    }

# dividers required to put the point where it belongs
dict_units_dividers = {
    "0": 1000,
     "1": 1000,
     "2": 100000,
     "3": 100000,
     "4": 100000,
     "5": 10000,
     "6": 10000,
     "7": 10000,
     "8": 100000,
     "9": 1,
     ".": 1,
    }

dict_typeConversions = {
    "TEXT": str,
    "DOUBLE": float,
    "DATE": complex,
    "LONG": int,
    }

REPLY_ACK = '?'
BEEP = 'BEEP/1\r\n'.encode('ascii')


class GSIPing(QThread):
    found_tachy = pyqtSignal(str)
    found_something = pyqtSignal(str)
    timed_out = pyqtSignal(str)

    def __init__(self, port_name, timeout=2000):
        self.ser = QSerialPort()
        self.ser.setPortName(port_name)
        self.ser.open(QSerialPort.ReadWrite)
        self.ser.write(BEEP)
        self.timer = QTimer()
        self.timer.setInterval(timeout)
        self.timer.setSingleShot(True)
        self.timer.timeout.connect(self.read)
        self.timer.start()
        super().__init__()

    def read(self):
        if self.ser.canReadLine():
            reply = bytes(self.ser.readLine()).decode('ascii')
            self.ser.close()
            if reply.startswith(REPLY_ACK):
                self.found_tachy.emit(self.ser.portName())
            else:
                self.found_something.emit(self.ser.portName())
        else:
            self.timed_out.emit(self.ser.portName())

        self.ser.close()
        self.quit()


class GSIMessage:
    def __init__(self, wi, value):
        pass


class GSIWord:
    def __init__(self, txt):
        txt = txt.strip().replace('*', '')
        header = txt[:7]
        self.data = header[7:]
        self.wi = header[:2]
        self.len = len(self.data)
        self.index_info = None if header[3] == '.' else header[3]
        self.input_mode = None if header[4] == '.' else header[4]
        self.unit = header[5]
        sign = header[6]
        self.multiplier = 1 if sign == '+' else -1

    def get_value(self):
        return int(self.data) / dict_units_dividers * self.sign

    def get_content(self):
        return dict_labels[self.wi]


def parse_word(word):
    header = word[:7]
    data = word[7:]
    wi = header


def parse(line):
    extracted = {}
    units = {}
    # Leica supports two different formats: GSI-8 and GSI-16.
    # The numbers represent the available precision for storing values.
    # GSI-16 is identified by prefixing the dataset with an asterisk character
    if len(line) <= 4:
        return extracted, units
    if line[0] == "*":
        precision = 16
        line = line[1:]
    else:
        precision = 8
    extracted['precision'] = precision
    for part in line.split():
        identifier = part[:2]
        idExtension = part[2]
        compensatorInfo = part[3]
        sourceInfo = part[4]
        unitInfo = part[5]
        signInfo = part[6]
        value = part[7:]

        # The first two places define the content of a part, which in turn defines the data type of the value
        label = dict_labels[identifier]
        baseType = dict_formats[identifier]
        units[label] = dict_units_attributes_digits[unitInfo]
        try:
            # strings are simple, the require no further treatment before passing on, so they are exempt from processing:
            if not baseType == "TEXT":
                # everything else has to be cast to another type:
                value = dict_typeConversions[baseType](value)
                # Datetimes are now considered done, numbers still have to be adjusted for sign and precision:
                if not baseType == "DATE":
                    # positive values are identified by a '+' sign at the 7th place of a word

                    if not signInfo == "+":
                        value *= -1
                    value /= dict_units_dividers[unitInfo]
                    
            # converted values are labeled and written to a dict
            extracted[label] = value
        except ValueError:
            pass
    return extracted, units


def make_vertex(line):
    vtx_data = parse(line)[0]
    if 'targetZ' not in list(vtx_data.keys()):
        raise ValueError('No z coordinate in: ' + line)
    label = vtx_data['pointId']
    x = vtx_data['targetX']
    y = vtx_data['targetY']
    z = vtx_data['targetZ']
    return x, y, z


if __name__ == "__main__":
    testLine = '*11....+0000000000000473 21.022+0000000039809400 22.022+0000000010859950 31..00+0000000000000609 81..00+0000000565385748 82..00+0000005924615105 83..00+0000000000005224 87..10+0000000000000000 \r\n'
    parsed, units = parse(testLine)
    print(list(parsed.keys()))
    print(parsed['precision'])
    print(parsed['targetX'], parsed['targetY'], parsed['targetZ'])
    print(units['targetX'], units['targetY'], units['targetZ'])
    print(parsed['pointId'])

    reply = "*88..10+0000000000000000"
    reply = "*13....+0000000000TCR405"
    print(parse(reply))


from enum import Enum


GRC_OK = 0                               # Function successfully completed.
GRC_UNDEFINED = 1                        # Unknown error, result unspecified.
GRC_IVPARAM = 2                          # Invalid parameter detected. Result unspecified.
GRC_IVRESULT = 3                         # Invalid result.
GRC_FATAL = 4                            # Fatal error.
GRC_NOT_IMPL = 5                         # Not implemented yet.
GRC_TIME_OUT = 6                         # Function execution timed out. Result unspecified.
GRC_SET_INCOMPL = 7                      # Parameter setup for subsystem is incomplete.
GRC_ABORT = 8                            # Function execution has been aborted.
GRC_NOMEMORY = 9                         # Fatal error - not enough memory.
GRC_NOTINIT = 10                         # Fatal error - subsystem not initialized.
GRC_SHUT_DOWN = 12                       # Subsystem is down.
GRC_SYSBUSY = 13                         # System busy/already in use of another process. Cannot execute function.
GRC_HWFAILURE = 14                       # Fatal error - hardware failure.
GRC_ABORT_APPL = 15                      # Execution of application has been aborted (SHIFT-ESC).
GRC_LOW_POWER = 16                       # Operation aborted - insufficient power supply level.
GRC_IVVERSION = 17                       # Invalid version of file, ...
GRC_BATT_EMPTY = 18                      # Battery empty
GRC_NO_EVENT = 20                        # no event pending.
GRC_OUT_OF_TEMP = 21                     # out of temperature range
GRC_INSTRUMENT_TILT = 22                 # instrument tilting out of range
GRC_COM_SETTING = 23                     # communication error
GRC_NO_ACTION = 24                       # GRC_TYPE Input 'do no action'
GRC_SLEEP_MODE = 25                      # Instr. run into the sleep mode
GRC_NOTOK = 26                           # Function not successfully completed.
GRC_NA = 27                              # Not available
GRC_OVERFLOW = 28                        # Overflow error
GRC_STOPPED = 29                         # System or subsystem has been stopped
GRC_ANG_ERROR = 257                      # Angles and Inclinations not valid
GRC_ANG_INCL_ERROR = 258                 # inclinations not valid
GRC_ANG_BAD_ACC = 259                    # value accuracies not reached
GRC_ANG_BAD_ANGLE_ACC = 260              # angle-accuracies not reached
GRC_ANG_BAD_INCLIN_ACC = 261             # inclination accuracies not reached
GRC_ANG_WRITE_PROTECTED = 266            # no write access allowed
GRC_ANG_OUT_OF_RANGE = 267               # value out of range
GRC_ANG_IR_OCCURED = 268                 # function aborted due to interrupt
GRC_ANG_HZ_MOVED = 269                   # hz moved during incline measurement
GRC_ANG_OS_ERROR = 270                   # troubles with operation system
GRC_ANG_DATA_ERROR = 271                 # overflow at parameter values
GRC_ANG_PEAK_CNT_UFL = 272               # too less peaks
GRC_ANG_TIME_OUT = 273                   # reading timeout
GRC_ANG_TOO_MANY_EXPOS = 274             # too many exposures wanted
GRC_ANG_PIX_CTRL_ERR = 275               # picture height out of range
GRC_ANG_MAX_POS_SKIP = 276               # positive exposure dynamic overflow
GRC_ANG_MAX_NEG_SKIP = 277               # negative exposure dynamic overflow
GRC_ANG_EXP_LIMIT = 278                  # exposure time overflow
GRC_ANG_UNDER_EXPOSURE = 279             # picture underexposured
GRC_ANG_OVER_EXPOSURE = 280              # picture overexposured
GRC_ANG_TMANY_PEAKS = 300                # too many peaks detected
GRC_ANG_TLESS_PEAKS = 301                # too less peaks detected
GRC_ANG_PEAK_TOO_SLIM = 302              # peak too slim
GRC_ANG_PEAK_TOO_WIDE = 303              # peak to wide
GRC_ANG_BAD_PEAKDIFF = 304               # bad peak difference
GRC_ANG_UNDER_EXP_PICT = 305             # too less peak amplitude
GRC_ANG_PEAKS_INHOMOGEN = 306            # inhomogeneous peak amplitudes
GRC_ANG_NO_DECOD_POSS = 307              # no peak decoding possible
GRC_ANG_UNSTABLE_DECOD = 308             # peak decoding not stable
GRC_ANG_TLESS_FPEAKS = 309               # too less valid finepeaks
GRC_ATA_NOT_READY = 512                  # ATR-System is not ready.
GRC_ATA_NO_RESULT = 513                  # Result isn't available yet.
GRC_ATA_SEVERAL_TARGETS = 514            # Several Targets detected.
GRC_ATA_BIG_SPOT = 515                   # Spot is too big for analyse.
GRC_ATA_BACKGROUND = 516                 # Background is too bright.
GRC_ATA_NO_TARGETS = 517                 # No targets detected.
GRC_ATA_NOT_ACCURAT = 518                # Accuracy worse than asked for.
GRC_ATA_SPOT_ON_EDGE = 519               # Spot is on the edge of the sensing area.
GRC_ATA_BLOOMING = 522                   # Blooming or spot on edge detected.
GRC_ATA_NOT_BUSY = 523                   # ATR isn't in a continuous mode.
GRC_ATA_STRANGE_LIGHT = 524              # Not the spot of the own target illuminator.
GRC_ATA_V24_FAIL = 525                   # Communication error to sensor (ATR).
GRC_ATA_DECODE_ERROR = 526               # Received Arguments cannot be decoded
GRC_ATA_HZ_FAIL = 527                    # No Spot detected in Hz-direction.
GRC_ATA_V_FAIL = 528                     # No Spot detected in V-direction.
GRC_ATA_HZ_STRANGE_L = 529               # Strange light in Hz-direction.
GRC_ATA_V_STRANGE_L = 530                # Strange light in V-direction.
GRC_ATA_SLDR_TRANSFER_PENDING = 531      # On multiple ATA_SLDR_OpenTransfer.
GRC_ATA_SLDR_TRANSFER_ILLEGAL = 532      # No ATA_SLDR_OpenTransfer happened.
GRC_ATA_SLDR_DATA_ERROR = 533            # Unexpected data format received.
GRC_ATA_SLDR_CHK_SUM_ERROR = 534         # Checksum error in transmitted data.
GRC_ATA_SLDR_ADDRESS_ERROR = 535         # Address out of valid range.
GRC_ATA_SLDR_INV_LOADFILE = 536          # Firmware file has invalid format.
GRC_ATA_SLDR_UNSUPPORTED = 537           # Current (loaded) firmware doesn't support upload.
GRC_ATA_PS_NOT_READY = 538               # PS-System is not ready.
GRC_ATA_ATR_SYSTEM_ERR = 539             # ATR system error
GRC_EDM_SYSTEM_ERR = 769                 # Fatal EDM sensor error. See for the exact reason the original EDM sensor error number. In the most cases a service problem.
GRC_EDM_INVALID_COMMAND = 770            # Invalid command or unknown command, see command syntax.
GRC_EDM_BOOM_ERR = 771                   # Boomerang error.
GRC_EDM_SIGN_LOW_ERR = 772               # Received signal to low, prism to far away, or natural barrier, bad environment, etc.
GRC_EDM_DIL_ERR = 773                    # obsolete
GRC_EDM_SIGN_HIGH_ERR = 774              # Received signal to strong, prism to near, stranger light effect.
GRC_EDM_TIMEOUT = 775                    # Timeout, measuring time exceeded (signal too weak, beam interrupted,..)
GRC_EDM_FLUKT_ERR = 776                  # to much turbulences or distractions
GRC_EDM_FMOT_ERR = 777                   # filter motor defective
GRC_EDM_DEV_NOT_INSTALLED = 778          # Device like EGL, DL is not installed.
GRC_EDM_NOT_FOUND = 779                  # Search result invalid. For the exact explanation, see in the description of the called function.
GRC_EDM_ERROR_RECEIVED = 780             # Communication ok, but an error reported from the EDM sensor.
GRC_EDM_MISSING_SRVPWD = 781             # No service password is set.
GRC_EDM_INVALID_ANSWER = 782             # Communication ok, but an unexpected answer received.
GRC_EDM_SEND_ERR = 783                   # Data send error, sending buffer is full.
GRC_EDM_RECEIVE_ERR = 784                # Data receive error, like parity buffer overflow.
GRC_EDM_INTERNAL_ERR = 785               # Internal EDM subsystem error.
GRC_EDM_BUSY = 786                       # Sensor is working already, abort current measuring first.
GRC_EDM_NO_MEASACTIVITY = 787            # No measurement activity started.
GRC_EDM_CHKSUM_ERR = 788                 # Calculated checksum, resp. received data wrong (only in binary communication mode possible).
GRC_EDM_INIT_OR_STOP_ERR = 789           # During start up or shut down phase an error occured. It is saved in the DEL buffer.
GRC_EDM_SRL_NOT_AVAILABLE = 790          # Red laser not available on this sensor HW.
GRC_EDM_MEAS_ABORTED = 791               # Measurement will be aborted (will be used for the laser security)
GRC_EDM_SLDR_TRANSFER_PENDING = 798      # Multiple OpenTransfer calls.
GRC_EDM_SLDR_TRANSFER_ILLEGAL = 799      # No open transfer happened.
GRC_EDM_SLDR_DATA_ERROR = 800            # Unexpected data format received.
GRC_EDM_SLDR_CHK_SUM_ERROR = 801         # Checksum error in transmitted data.
GRC_EDM_SLDR_ADDR_ERROR = 802            # Address out of valid range.
GRC_EDM_SLDR_INV_LOADFILE = 803          # Firmware file has invalid format.
GRC_EDM_SLDR_UNSUPPORTED = 804           # Current (loaded) firmware doesn't support upload.
GRC_EDM_UNKNOW_ERR = 808                 # Undocumented error from the EDM sensor, should not occur.
GRC_EDM_DISTRANGE_ERR = 818              # Out of distance range (dist too small or large)
GRC_EDM_SIGNTONOISE_ERR = 819            # Signal to noise ratio too small
GRC_EDM_NOISEHIGH_ERR = 820              # Noise to high
GRC_EDM_PWD_NOTSET = 821                 # Password is not set
GRC_EDM_ACTION_NO_MORE_VALID = 822       # Elapsed time between prepare und start fast measurement for ATR to long
GRC_EDM_MULTRG_ERR = 823                 # Possibly more than one target (also a sensor error)
GRC_TMC_NO_FULL_CORRECTION = 1283        # Warning: measurement without full correction
GRC_TMC_ACCURACY_GUARANTEE = 1284        # Info: accuracy can not be guarantee
GRC_TMC_ANGLE_OK = 1285                  # Warning: only angle measurement valid
GRC_TMC_ANGLE_NOT_FULL_CORR = 1288       # Warning: only angle measurement valid but without full correction
GRC_TMC_ANGLE_NO_ACC_GUARANTY = 1289     # Info: only angle measurement valid but accuracy can not be guarantee
GRC_TMC_ANGLE_ERROR = 1290               # Error: no angle measurement
GRC_TMC_DIST_PPM = 1291                  # Error: wrong setting of PPM or MM on EDM
GRC_TMC_DIST_ERROR = 1292                # Error: distance measurement not done (no aim, etc.)
GRC_TMC_BUSY = 1293                      # Error: system is busy (no measurement done)
GRC_TMC_SIGNAL_ERROR = 1294              # Error: no signal on EDM (only in signal mode)
MOT_RC_UNREADY = 1792                    # Motorization not ready
MOT_RC_BUSY = 1793                       # Motorization is handling another task
MOT_RC_NOT_OCONST = 1794                 # Not in velocity mode
MOT_RC_NOT_CONFIG = 1795                 # Motorization is in the wrong mode or busy
MOT_RC_NOT_POSIT = 1796                  # Not in posit mode
MOT_RC_NOT_SERVICE = 1797                # Not in service mode
MOT_RC_NOT_BUSY = 1798                   # Motorization is handling no task
MOT_RC_NOT_LOCK = 1799                   # Not in tracking mode
GRC_BMM_XFER_PENDING = 2305              # Loading process already opened
GRC_BMM_NO_XFER_OPEN = 2306              # Transfer not opened
GRC_BMM_UNKNOWN_CHARSET = 2307           # Unknown character set
GRC_BMM_NOT_INSTALLED = 2308             # Display module not present
GRC_BMM_ALREADY_EXIST = 2309             # Character set already exists
GRC_BMM_CANT_DELETE = 2310               # Character set cannot be deleted
GRC_BMM_MEM_ERROR = 2311                 # Memory cannot be allocated
GRC_BMM_CHARSET_USED = 2312              # Character set still used
GRC_BMM_CHARSET_SAVED = 2313             # Charset cannot be deleted or is protected
GRC_BMM_INVALID_ADR = 2314               # Attempt to copy a character block outside the allocated memory
GRC_BMM_CANCELANDADR_ERROR = 2315        # Error during release of allocated memory
GRC_BMM_INVALID_SIZE = 2316              # Number of bytes specified in header does not match the bytes read
GRC_BMM_CANCELANDINVSIZE_ERROR = 2317    # Allocated memory could not be released
GRC_BMM_ALL_GROUP_OCC = 2318             # Max. number of character sets already loaded
GRC_BMM_CANT_DEL_LAYERS = 2319           # Layer cannot be deleted
GRC_BMM_UNKNOWN_LAYER = 2320             # Required layer does not exist
GRC_BMM_INVALID_LAYERLEN = 2321          # Layer length exceeds maximum
GRC_COM_ERO = 3072                       # Initiate Extended Runtime Operation (ERO).
GRC_COM_CANT_ENCODE = 3073               # Cannot encode arguments in client.
GRC_COM_CANT_DECODE = 3074               # Cannot decode results in client.
GRC_COM_CANT_SEND = 3075                 # Hardware error while sending.
GRC_COM_CANT_RECV = 3076                 # Hardware error while receiving.
GRC_COM_TIMEDOUT = 3077                  # Request timed out.
GRC_COM_WRONG_FORMAT = 3078              # Packet format error.
GRC_COM_VER_MISMATCH = 3079              # Version mismatch between client and server.
GRC_COM_CANT_DECODE_REQ = 3080           # Cannot decode arguments in server.
GRC_COM_PROC_UNAVAIL = 3081              # Unknown RPC, procedure ID invalid.
GRC_COM_CANT_ENCODE_REP = 3082           # Cannot encode results in server.
GRC_COM_SYSTEM_ERR = 3083                # Unspecified generic system error.
GRC_COM_FAILED = 3085                    # Unspecified error.
GRC_COM_NO_BINARY = 3086                 # Binary protocol not available.
GRC_COM_INTR = 3087                      # Call interrupted.
GRC_COM_REQUIRES_8DBITS = 3090           # Protocol needs 8bit encoded characters.
GRC_COM_TR_ID_MISMATCH = 3093            # TRANSACTIONS ID mismatch error.
GRC_COM_NOT_GEOCOM = 3094                # Protocol not recognizable.
GRC_COM_UNKNOWN_PORT = 3095              # (WIN) Invalid port address.
GRC_COM_ERO_END = 3099                   # ERO is terminating.
GRC_COM_OVERRUN = 3100                   # Internal error: data buffer overflow.
GRC_COM_SRVR_RX_CHECKSUM_ERRR = 3101     # Invalid checksum on server side received.
GRC_COM_CLNT_RX_CHECKSUM_ERRR = 3102     # Invalid checksum on client side received.
GRC_COM_PORT_NOT_AVAILABLE = 3103        # (WIN) Port not available.
GRC_COM_PORT_NOT_OPEN = 3104             # (WIN) Port not opened.
GRC_COM_NO_PARTNER = 3105                # (WIN) Unable to find TPS.
GRC_COM_ERO_NOT_STARTED = 3106           # Extended Runtime Operation could not be started.
GRC_COM_CONS_REQ = 3107                  # Att to send cons reqs
GRC_COM_SRVR_IS_SLEEPING = 3108          # TPS has gone to sleep. Wait and try again.
GRC_COM_SRVR_IS_OFF = 3109               # TPS has shut down. Wait and try again.
WIR_PTNR_OVERFLOW = 5121                 # point number overflow
WIR_NUM_ASCII_CARRY = 5122               # carry from number to ASCII conversion
WIR_PTNR_NO_INC = 5123                   # can't increment point number
WIR_STEP_SIZE = 5124                     # wrong step size
WIR_BUSY = 5125                          # resource occupied
WIR_CONFIG_FNC = 5127                    # user function selected
WIR_CANT_OPEN_FILE = 5128                # can't open file
WIR_FILE_WRITE_ERROR = 5129              # can't write into file
WIR_MEDIUM_NOMEM = 5130                  # no anymore memory on PC-Card
WIR_NO_MEDIUM = 5131                     # no PC-Card
WIR_EMPTY_FILE = 5132                    # empty GSI file
WIR_INVALID_DATA = 5133                  # invalid data in GSI file
WIR_F2_BUTTON = 5134                     # F2 button pressed
WIR_F3_BUTTON = 5135                     # F3 button pressed
WIR_F4_BUTTON = 5136                     # F4 button pressed
WIR_SHF2_BUTTON = 5137                   # SHIFT F2 button pressed
GRC_AUT_TIMEOUT = 8704                   # Position not reached
GRC_AUT_DETENT_ERROR = 8705              # Positioning not possible due to mounted EDM
GRC_AUT_ANGLE_ERROR = 8706               # Angle measurement error
GRC_AUT_MOTOR_ERROR = 8707               # Motorisation error
GRC_AUT_INCACC = 8708                    # Position not exactly reached
GRC_AUT_DEV_ERROR = 8709                 # Deviation measurement error
GRC_AUT_NO_TARGET = 8710                 # No target detected
GRC_AUT_MULTIPLE_TARGETS = 8711          # Multiple target detected
GRC_AUT_BAD_ENVIRONMENT = 8712           # Bad environment conditions
GRC_AUT_DETECTOR_ERROR = 8713            # Error in target acquisition
GRC_AUT_NOT_ENABLED = 8714               # Target acquisition not enabled
GRC_AUT_CALACC = 8715                    # ATR-Calibration failed
GRC_AUT_ACCURACY = 8716                  # Target position not exactly reached
GRC_AUT_DIST_STARTED = 8717              # Info: dist. measurement has been started
GRC_AUT_SUPPLY_TOO_HIGH = 8718           # external Supply voltage is too high
GRC_AUT_SUPPLY_TOO_LOW = 8719            # int. or ext. Supply voltage is too low
GRC_AUT_NO_WORKING_AREA = 8720           # working area not set
GRC_AUT_ARRAY_FULL = 8721                # power search data array is filled
GRC_AUT_NO_DATA = 8722                   # no data available
BAP_CHANGE_ALL_TO_DIST = 9217            # Command changed from ALL to DIST
GRC_KDM_NOT_AVAILABLE = 12544            # KDM device is not available.
GRC_FTR_FILEACCESS = 13056               # File access error
GRC_FTR_WRONGFILEBLOCKNUMBER = 13057     # block number was not the expected one
GRC_FTR_NOTENOUGHSPACE = 13058           # not enough space on device to proceed uploading
GRC_FTR_INVALIDINPUT = 13059             # Rename of file failed.
GRC_FTR_MISSINGSETUP = 13060             # invalid parameter as input


MESSAGES = {0: "Function successfully completed.",
            1: "Unknown error, result unspecified.",
            2: "Invalid parameter detected. Result unspecified.",
            3: "Invalid result.",
            4: "Fatal error.",
            5: "Not implemented yet.",
            6: "Function execution timed out. Result unspecified.",
            7: "Parameter setup for subsystem is incomplete.",
            8: "Function execution has been aborted.",
            9: "Fatal error - not enough memory.",
            10: "Fatal error - subsystem not initialized.",
            12: "Subsystem is down.",
            13: "System busy/already in use of another process. Cannot execute function.",
            14: "Fatal error - hardware failure.",
            15: "Execution of application has been aborted (SHIFT-ESC).",
            16: "Operation aborted - insufficient power supply level.",
            17: "Invalid version of file, ...",
            18: "Battery empty",
            20: "no event pending.",
            21: "out of temperature range",
            22: "instrument tilting out of range",
            23: "communication error",
            24: "GRC_TYPE Input 'do no action'",
            25: "Instr. run into the sleep mode",
            26: "Function not successfully completed.",
            27: "Not available",
            28: "Overflow error",
            29: "System or subsystem has been stopped",
            257: "Angles and Inclinations not valid",
            258: "inclinations not valid",
            259: "value accuracies not reached",
            260: "angle-accuracies not reached",
            261: "inclination accuracies not reached",
            266: "no write access allowed",
            267: "value out of range",
            268: "function aborted due to interrupt",
            269: "hz moved during incline measurement",
            270: "troubles with operation system",
            271: "overflow at parameter values",
            272: "too less peaks",
            273: "reading timeout",
            274: "too many exposures wanted",
            275: "picture height out of range",
            276: "positive exposure dynamic overflow",
            277: "negative exposure dynamic overflow",
            278: "exposure time overflow",
            279: "picture underexposured",
            280: "picture overexposured",
            300: "too many peaks detected",
            301: "too less peaks detected",
            302: "peak too slim",
            303: "peak to wide",
            304: "bad peak difference",
            305: "too less peak amplitude",
            306: "inhomogeneous peak amplitudes",
            307: "no peak decoding possible",
            308: "peak decoding not stable",
            309: "too less valid finepeaks",
            512: "ATR-System is not ready.",
            513: "Result isn't available yet.",
            514: "Several Targets detected.",
            515: "Spot is too big for analyse.",
            516: "Background is too bright.",
            517: "No targets detected.",
            518: "Accuracy worse than asked for.",
            519: "Spot is on the edge of the sensing area.",
            522: "Blooming or spot on edge detected.",
            523: "ATR isn't in a continuous mode.",
            524: "Not the spot of the own target illuminator.",
            525: "Communication error to sensor (ATR).",
            526: "Received Arguments cannot be decoded",
            527: "No Spot detected in Hz-direction.",
            528: "No Spot detected in V-direction.",
            529: "Strange light in Hz-direction.",
            530: "Strange light in V-direction.",
            531: "On multiple ATA_SLDR_OpenTransfer.",
            532: "No ATA_SLDR_OpenTransfer happened.",
            533: "Unexpected data format received.",
            534: "Checksum error in transmitted data.",
            535: "Address out of valid range.",
            536: "Firmware file has invalid format.",
            537: "Current (loaded) firmware doesn't support upload.",
            538: "PS-System is not ready.",
            539: "ATR system error",
            769: "Fatal EDM sensor error. See for the exact reason the original EDM sensor error number. In the most cases a service problem.",
            770: "Invalid command or unknown command, see command syntax.",
            771: "Boomerang error.",
            772: "Received signal to low, prism to far away, or natural barrier, bad environment, etc.",
            773: "obsolete",
            774: "Received signal to strong, prism to near, stranger light effect.",
            775: "Timeout, measuring time exceeded (signal too weak, beam interrupted,..)",
            776: "to much turbulences or distractions",
            777: "filter motor defective",
            778: "Device like EGL, DL is not installed.",
            779: "Search result invalid. For the exact explanation, see in the description of the called function.",
            780: "Communication ok, but an error reported from the EDM sensor.",
            781: "No service password is set.",
            782: "Communication ok, but an unexpected answer received.",
            783: "Data send error, sending buffer is full.",
            784: "Data receive error, like parity buffer overflow.",
            785: "Internal EDM subsystem error.",
            786: "Sensor is working already, abort current measuring first.",
            787: "No measurement activity started.",
            788: "Calculated checksum, resp. received data wrong (only in binary communication mode possible).",
            789: "During start up or shut down phase an error occured. It is saved in the DEL buffer.",
            790: "Red laser not available on this sensor HW.",
            791: "Measurement will be aborted (will be used for the laser security)",
            798: "Multiple OpenTransfer calls.",
            799: "No open transfer happened.",
            800: "Unexpected data format received.",
            801: "Checksum error in transmitted data.",
            802: "Address out of valid range.",
            803: "Firmware file has invalid format.",
            804: "Current (loaded) firmware doesn't support upload.",
            808: "Undocumented error from the EDM sensor, should not occur.",
            818: "Out of distance range (dist too small or large)",
            819: "Signal to noise ratio too small",
            820: "Noise to high",
            821: "Password is not set",
            822: "Elapsed time between prepare und start fast measurement for ATR to long",
            823: "Possibly more than one target (also a sensor error)",
            1283: "Warning: measurement without full correction",
            1284: "Info: accuracy can not be guarantee",
            1285: "Warning: only angle measurement valid",
            1288: "Warning: only angle measurement valid but without full correction",
            1289: "Info: only angle measurement valid but accuracy can not be guarantee",
            1290: "Error: no angle measurement",
            1291: "Error: wrong setting of PPM or MM on EDM",
            1292: "Error: distance measurement not done (no aim, etc.)",
            1293: "Error: system is busy (no measurement done)",
            1294: "Error: no signal on EDM (only in signal mode)",
            1792: "Motorization not ready",
            1793: "Motorization is handling another task",
            1794: "Not in velocity mode",
            1795: "Motorization is in the wrong mode or busy",
            1796: "Not in posit mode",
            1797: "Not in service mode",
            1798: "Motorization is handling no task",
            1799: "Not in tracking mode",
            2305: "Loading process already opened",
            2306: "Transfer not opened",
            2307: "Unknown character set",
            2308: "Display module not present",
            2309: "Character set already exists",
            2310: "Character set cannot be deleted",
            2311: "Memory cannot be allocated",
            2312: "Character set still used",
            2313: "Charset cannot be deleted or is protected",
            2314: "Attempt to copy a character block outside the allocated memory",
            2315: "Error during release of allocated memory",
            2316: "Number of bytes specified in header does not match the bytes read",
            2317: "Allocated memory could not be released",
            2318: "Max. number of character sets already loaded",
            2319: "Layer cannot be deleted",
            2320: "Required layer does not exist",
            2321: "Layer length exceeds maximum",
            3072: "Initiate Extended Runtime Operation (ERO).",
            3073: "Cannot encode arguments in client.",
            3074: "Cannot decode results in client.",
            3075: "Hardware error while sending.",
            3076: "Hardware error while receiving.",
            3077: "Request timed out.",
            3078: "Packet format error.",
            3079: "Version mismatch between client and server.",
            3080: "Cannot decode arguments in server.",
            3081: "Unknown RPC, procedure ID invalid.",
            3082: "Cannot encode results in server.",
            3083: "Unspecified generic system error.",
            3085: "Unspecified error.",
            3086: "Binary protocol not available.",
            3087: "Call interrupted.",
            3090: "Protocol needs 8bit encoded characters.",
            3093: "TRANSACTIONS ID mismatch error.",
            3094: "Protocol not recognizable.",
            3095: "(WIN) Invalid port address.",
            3099: "ERO is terminating.",
            3100: "Internal error: data buffer overflow.",
            3101: "Invalid checksum on server side received.",
            3102: "Invalid checksum on client side received.",
            3103: "(WIN) Port not available.",
            3104: "(WIN) Port not opened.",
            3105: "(WIN) Unable to find TPS.",
            3106: "Extended Runtime Operation could not be started.",
            3107: "Att to send cons reqs",
            3108: "TPS has gone to sleep. Wait and try again.",
            3109: "TPS has shut down. Wait and try again.",
            5121: "point number overflow",
            5122: "carry from number to ASCII conversion",
            5123: "can't increment point number",
            5124: "wrong step size",
            5125: "resource occupied",
            5127: "user function selected",
            5128: "can't open file",
            5129: "can't write into file",
            5130: "no anymore memory on PC-Card",
            5131: "no PC-Card",
            5132: "empty GSI file",
            5133: "invalid data in GSI file",
            5134: "F2 button pressed",
            5135: "F3 button pressed",
            5136: "F4 button pressed",
            5137: "SHIFT F2 button pressed",
            8704: "Position not reached",
            8705: "Positioning not possible due to mounted EDM",
            8706: "Angle measurement error",
            8707: "Motorisation error",
            8708: "Position not exactly reached",
            8709: "Deviation measurement error",
            8710: "No target detected",
            8711: "Multiple target detected",
            8712: "Bad environment conditions",
            8713: "Error in target acquisition",
            8714: "Target acquisition not enabled",
            8715: "ATR-Calibration failed",
            8716: "Target position not exactly reached",
            8717: "Info: dist. measurement has been started",
            8718: "external Supply voltage is too high",
            8719: "int. or ext. Supply voltage is too low",
            8720: "working area not set",
            8721: "power search data array is filled",
            8722: "no data available",
            9217: "Command changed from ALL to DIST",
            12544: "KDM device is not available.",
            13056: "File access error",
            13057: "block number was not the expected one",
            13058: "not enough space on device to proceed uploading",
            13059: "Rename of file failed.",
            13060: "invalid parameter as input",
            }


CODES = {0: "GRC_OK",
         1: "GRC_UNDEFINED",
         2: "GRC_IVPARAM",
         3: "GRC_IVRESULT",
         4: "GRC_FATAL",
         5: "GRC_NOT_IMPL",
         6: "GRC_TIME_OUT",
         7: "GRC_SET_INCOMPL",
         8: "GRC_ABORT",
         9: "GRC_NOMEMORY",
         10: "GRC_NOTINIT",
         12: "GRC_SHUT_DOWN",
         13: "GRC_SYSBUSY",
         14: "GRC_HWFAILURE",
         15: "GRC_ABORT_APPL",
         16: "GRC_LOW_POWER",
         17: "GRC_IVVERSION",
         18: "GRC_BATT_EMPTY",
         20: "GRC_NO_EVENT",
         21: "GRC_OUT_OF_TEMP",
         22: "GRC_INSTRUMENT_TILT",
         23: "GRC_COM_SETTING",
         24: "GRC_NO_ACTION",
         25: "GRC_SLEEP_MODE",
         26: "GRC_NOTOK",
         27: "GRC_NA",
         28: "GRC_OVERFLOW",
         29: "GRC_STOPPED",
         257: "GRC_ANG_ERROR",
         258: "GRC_ANG_INCL_ERROR",
         259: "GRC_ANG_BAD_ACC",
         260: "GRC_ANG_BAD_ANGLE_ACC",
         261: "GRC_ANG_BAD_INCLIN_ACC",
         266: "GRC_ANG_WRITE_PROTECTED",
         267: "GRC_ANG_OUT_OF_RANGE",
         268: "GRC_ANG_IR_OCCURED",
         269: "GRC_ANG_HZ_MOVED",
         270: "GRC_ANG_OS_ERROR",
         271: "GRC_ANG_DATA_ERROR",
         272: "GRC_ANG_PEAK_CNT_UFL",
         273: "GRC_ANG_TIME_OUT",
         274: "GRC_ANG_TOO_MANY_EXPOS",
         275: "GRC_ANG_PIX_CTRL_ERR",
         276: "GRC_ANG_MAX_POS_SKIP",
         277: "GRC_ANG_MAX_NEG_SKIP",
         278: "GRC_ANG_EXP_LIMIT",
         279: "GRC_ANG_UNDER_EXPOSURE",
         280: "GRC_ANG_OVER_EXPOSURE",
         300: "GRC_ANG_TMANY_PEAKS",
         301: "GRC_ANG_TLESS_PEAKS",
         302: "GRC_ANG_PEAK_TOO_SLIM",
         303: "GRC_ANG_PEAK_TOO_WIDE",
         304: "GRC_ANG_BAD_PEAKDIFF",
         305: "GRC_ANG_UNDER_EXP_PICT",
         306: "GRC_ANG_PEAKS_INHOMOGEN",
         307: "GRC_ANG_NO_DECOD_POSS",
         308: "GRC_ANG_UNSTABLE_DECOD",
         309: "GRC_ANG_TLESS_FPEAKS",
         512: "GRC_ATA_NOT_READY",
         513: "GRC_ATA_NO_RESULT",
         514: "GRC_ATA_SEVERAL_TARGETS",
         515: "GRC_ATA_BIG_SPOT",
         516: "GRC_ATA_BACKGROUND",
         517: "GRC_ATA_NO_TARGETS",
         518: "GRC_ATA_NOT_ACCURAT",
         519: "GRC_ATA_SPOT_ON_EDGE",
         522: "GRC_ATA_BLOOMING",
         523: "GRC_ATA_NOT_BUSY",
         524: "GRC_ATA_STRANGE_LIGHT",
         525: "GRC_ATA_V24_FAIL",
         526: "GRC_ATA_DECODE_ERROR",
         527: "GRC_ATA_HZ_FAIL",
         528: "GRC_ATA_V_FAIL",
         529: "GRC_ATA_HZ_STRANGE_L",
         530: "GRC_ATA_V_STRANGE_L",
         531: "GRC_ATA_SLDR_TRANSFER_PENDING",
         532: "GRC_ATA_SLDR_TRANSFER_ILLEGAL",
         533: "GRC_ATA_SLDR_DATA_ERROR",
         534: "GRC_ATA_SLDR_CHK_SUM_ERROR",
         535: "GRC_ATA_SLDR_ADDRESS_ERROR",
         536: "GRC_ATA_SLDR_INV_LOADFILE",
         537: "GRC_ATA_SLDR_UNSUPPORTED",
         538: "GRC_ATA_PS_NOT_READY",
         539: "GRC_ATA_ATR_SYSTEM_ERR",
         769: "GRC_EDM_SYSTEM_ERR",
         770: "GRC_EDM_INVALID_COMMAND",
         771: "GRC_EDM_BOOM_ERR",
         772: "GRC_EDM_SIGN_LOW_ERR",
         773: "GRC_EDM_DIL_ERR",
         774: "GRC_EDM_SIGN_HIGH_ERR",
         775: "GRC_EDM_TIMEOUT",
         776: "GRC_EDM_FLUKT_ERR",
         777: "GRC_EDM_FMOT_ERR",
         778: "GRC_EDM_DEV_NOT_INSTALLED",
         779: "GRC_EDM_NOT_FOUND",
         780: "GRC_EDM_ERROR_RECEIVED",
         781: "GRC_EDM_MISSING_SRVPWD",
         782: "GRC_EDM_INVALID_ANSWER",
         783: "GRC_EDM_SEND_ERR",
         784: "GRC_EDM_RECEIVE_ERR",
         785: "GRC_EDM_INTERNAL_ERR",
         786: "GRC_EDM_BUSY",
         787: "GRC_EDM_NO_MEASACTIVITY",
         788: "GRC_EDM_CHKSUM_ERR",
         789: "GRC_EDM_INIT_OR_STOP_ERR",
         790: "GRC_EDM_SRL_NOT_AVAILABLE",
         791: "GRC_EDM_MEAS_ABORTED",
         798: "GRC_EDM_SLDR_TRANSFER_PENDING",
         799: "GRC_EDM_SLDR_TRANSFER_ILLEGAL",
         800: "GRC_EDM_SLDR_DATA_ERROR",
         801: "GRC_EDM_SLDR_CHK_SUM_ERROR",
         802: "GRC_EDM_SLDR_ADDR_ERROR",
         803: "GRC_EDM_SLDR_INV_LOADFILE",
         804: "GRC_EDM_SLDR_UNSUPPORTED",
         808: "GRC_EDM_UNKNOW_ERR",
         818: "GRC_EDM_DISTRANGE_ERR",
         819: "GRC_EDM_SIGNTONOISE_ERR",
         820: "GRC_EDM_NOISEHIGH_ERR",
         821: "GRC_EDM_PWD_NOTSET",
         822: "GRC_EDM_ACTION_NO_MORE_VALID",
         823: "GRC_EDM_MULTRG_ERR",
         1283: "GRC_TMC_NO_FULL_CORRECTION",
         1284: "GRC_TMC_ACCURACY_GUARANTEE",
         1285: "GRC_TMC_ANGLE_OK",
         1288: "GRC_TMC_ANGLE_NOT_FULL_CORR",
         1289: "GRC_TMC_ANGLE_NO_ACC_GUARANTY",
         1290: "GRC_TMC_ANGLE_ERROR",
         1291: "GRC_TMC_DIST_PPM",
         1292: "GRC_TMC_DIST_ERROR",
         1293: "GRC_TMC_BUSY",
         1294: "GRC_TMC_SIGNAL_ERROR",
         1792: "MOT_RC_UNREADY",
         1793: "MOT_RC_BUSY",
         1794: "MOT_RC_NOT_OCONST",
         1795: "MOT_RC_NOT_CONFIG",
         1796: "MOT_RC_NOT_POSIT",
         1797: "MOT_RC_NOT_SERVICE",
         1798: "MOT_RC_NOT_BUSY",
         1799: "MOT_RC_NOT_LOCK",
         2305: "GRC_BMM_XFER_PENDING",
         2306: "GRC_BMM_NO_XFER_OPEN",
         2307: "GRC_BMM_UNKNOWN_CHARSET",
         2308: "GRC_BMM_NOT_INSTALLED",
         2309: "GRC_BMM_ALREADY_EXIST",
         2310: "GRC_BMM_CANT_DELETE",
         2311: "GRC_BMM_MEM_ERROR",
         2312: "GRC_BMM_CHARSET_USED",
         2313: "GRC_BMM_CHARSET_SAVED",
         2314: "GRC_BMM_INVALID_ADR",
         2315: "GRC_BMM_CANCELANDADR_ERROR",
         2316: "GRC_BMM_INVALID_SIZE",
         2317: "GRC_BMM_CANCELANDINVSIZE_ERROR",
         2318: "GRC_BMM_ALL_GROUP_OCC",
         2319: "GRC_BMM_CANT_DEL_LAYERS",
         2320: "GRC_BMM_UNKNOWN_LAYER",
         2321: "GRC_BMM_INVALID_LAYERLEN",
         3072: "GRC_COM_ERO",
         3073: "GRC_COM_CANT_ENCODE",
         3074: "GRC_COM_CANT_DECODE",
         3075: "GRC_COM_CANT_SEND",
         3076: "GRC_COM_CANT_RECV",
         3077: "GRC_COM_TIMEDOUT",
         3078: "GRC_COM_WRONG_FORMAT",
         3079: "GRC_COM_VER_MISMATCH",
         3080: "GRC_COM_CANT_DECODE_REQ",
         3081: "GRC_COM_PROC_UNAVAIL",
         3082: "GRC_COM_CANT_ENCODE_REP",
         3083: "GRC_COM_SYSTEM_ERR",
         3085: "GRC_COM_FAILED",
         3086: "GRC_COM_NO_BINARY",
         3087: "GRC_COM_INTR",
         3090: "GRC_COM_REQUIRES_8DBITS",
         3093: "GRC_COM_TR_ID_MISMATCH",
         3094: "GRC_COM_NOT_GEOCOM",
         3095: "GRC_COM_UNKNOWN_PORT",
         3099: "GRC_COM_ERO_END",
         3100: "GRC_COM_OVERRUN",
         3101: "GRC_COM_SRVR_RX_CHECKSUM_ERRR",
         3102: "GRC_COM_CLNT_RX_CHECKSUM_ERRR",
         3103: "GRC_COM_PORT_NOT_AVAILABLE",
         3104: "GRC_COM_PORT_NOT_OPEN",
         3105: "GRC_COM_NO_PARTNER",
         3106: "GRC_COM_ERO_NOT_STARTED",
         3107: "GRC_COM_CONS_REQ",
         3108: "GRC_COM_SRVR_IS_SLEEPING",
         3109: "GRC_COM_SRVR_IS_OFF",
         5121: "WIR_PTNR_OVERFLOW",
         5122: "WIR_NUM_ASCII_CARRY",
         5123: "WIR_PTNR_NO_INC",
         5124: "WIR_STEP_SIZE",
         5125: "WIR_BUSY",
         5127: "WIR_CONFIG_FNC",
         5128: "WIR_CANT_OPEN_FILE",
         5129: "WIR_FILE_WRITE_ERROR",
         5130: "WIR_MEDIUM_NOMEM",
         5131: "WIR_NO_MEDIUM",
         5132: "WIR_EMPTY_FILE",
         5133: "WIR_INVALID_DATA",
         5134: "WIR_F2_BUTTON",
         5135: "WIR_F3_BUTTON",
         5136: "WIR_F4_BUTTON",
         5137: "WIR_SHF2_BUTTON",
         8704: "GRC_AUT_TIMEOUT",
         8705: "GRC_AUT_DETENT_ERROR",
         8706: "GRC_AUT_ANGLE_ERROR",
         8707: "GRC_AUT_MOTOR_ERROR",
         8708: "GRC_AUT_INCACC",
         8709: "GRC_AUT_DEV_ERROR",
         8710: "GRC_AUT_NO_TARGET",
         8711: "GRC_AUT_MULTIPLE_TARGETS",
         8712: "GRC_AUT_BAD_ENVIRONMENT",
         8713: "GRC_AUT_DETECTOR_ERROR",
         8714: "GRC_AUT_NOT_ENABLED",
         8715: "GRC_AUT_CALACC",
         8716: "GRC_AUT_ACCURACY",
         8717: "GRC_AUT_DIST_STARTED",
         8718: "GRC_AUT_SUPPLY_TOO_HIGH",
         8719: "GRC_AUT_SUPPLY_TOO_LOW",
         8720: "GRC_AUT_NO_WORKING_AREA",
         8721: "GRC_AUT_ARRAY_FULL",
         8722: "GRC_AUT_NO_DATA",
         9217: "BAP_CHANGE_ALL_TO_DIST",
         12544: "GRC_KDM_NOT_AVAILABLE",
         13056: "GRC_FTR_FILEACCESS",
         13057: "GRC_FTR_WRONGFILEBLOCKNUMBER",
         13058: "GRC_FTR_NOTENOUGHSPACE",
         13059: "GRC_FTR_INVALIDINPUT",
         13060: "GRC_FTR_MISSINGSETUP",
         }

AUS_GetRcsSearchSwitch = 18010
AUS_GetUserAtrState = 18006
AUS_GetUserLockState = 18008
AUS_SetUserAtrState = 18005
AUS_SetUserLockState = 18007
AUS_SwitchRcsSearch = 18009
AUT_ChangeFace = 9028
AUT_ChangeFace4 = 9028
AUT_FineAdjust = 9037
AUT_FineAdjust3 = 9037
AUT_GetATRStatus = 9019
AUT_GetFineAdjustMode = 9030
AUT_GetLockStatus = 9021
AUT_GetSearchArea = 9042
AUT_GetUserSpiral = 9040
AUT_LockIn = 9013
AUT_MakePositioning = 9027
AUT_MakePositioning4 = 9027
AUT_PS_EnableRange = 9048
AUT_PS_SearchNext = 9051
AUT_PS_SearchWindow = 9052
AUT_PS_SetRange = 9047
AUT_ReadTimeout = 9012
AUT_ReadTol = 9008
AUT_Search = 9029
AUT_Search2 = 9029
AUT_SetATRStatus = 9018
AUT_SetFineAdjustMode = 9031
AUT_SetLockStatus = 9020
AUT_SetSearchArea = 9043
AUT_SetTimeout = 9011
AUT_SetTol = 9007
AUT_SetUserSpiral = 9041
BAP_GetLastDisplayedError = 17003
BAP_GetMeasPrg = 17018
BAP_GetPrismDef = 17023
BAP_GetPrismType = 17009
BAP_GetTargetType = 17022
BAP_MeasDistanceAngle = 17017
BAP_SearchTarget = 17020
BAP_SetMeasPrg = 17019
BAP_SetPrismDef = 17024
BAP_SetPrismType = 17008
BAP_SetTargetType = 17021
BMM_BeepAlarm = 11004
BMM_BeepNormal = 11003
BMM_BeepOff = 11002
BMM_BeepOn = 11001
COM_EnableSignOff = 115
COM_GetBinaryAvailable = 113
COM_GetDoublePrecision = 108
COM_GetSWVersion = 110
COM_Local = 1
COM_NullProc = 0
COM_SetBinaryAvailable = 114
COM_SetDoublePrecision = 107
COM_SetSendDelay = 109
COM_SwitchOffTPS = 112
COM_SwitchOnTPS = 111
CSV_GetDateTime = 5008
CSV_GetDeviceConfig = 5035
CSV_GetInstrumentName = 5004
CSV_GetInstrumentNo = 5003
CSV_GetIntTemp = 5011
CSV_GetSWVersion = 5034
CSV_GetSWVersion2 = 5034
CSV_GetUserInstrumentName = 5006
CSV_GetVBat = 5009
CSV_GetVMem = 5010
CSV_SetDateTime = 5007
CSV_SetUserInstrumentName = 5005
CTL_GetUpCounter = 12003
EDM_GetBumerang = 1044
EDM_GetEglIntensity = 1058
EDM_GetTrkLightBrightness = 1041
EDM_GetTrkLightSwitch = 1040
EDM_Laserpointer = 1004
EDM_On = 1010
EDM_SetBumerang = 1007
EDM_SetEglIntensity = 1059
EDM_SetTrkLightBrightness = 1032
EDM_SetTrkLightSwitch = 1031
IOS_BeepOff = 20000
IOS_BeepOn = 20001
MOT_ReadLockStatus = 6021
MOT_SetVelocity = 6004
MOT_StartController = 6001
MOT_StopController = 6002
SUP_GetConfig = 14001
SUP_SetConfig = 14002
SUP_SwitchLowTempControl = 14003
TMC_DoMeasure = 2008
TMC_GetAngSwitch = 2014
TMC_GetAngle1 = 2003
TMC_GetAngle5 = 2107
TMC_GetAtmCorr = 2029
TMC_GetCoordinate = 2082
TMC_GetCoordinate1 = 2082
TMC_GetEdmMode = 2021
TMC_GetFace = 2026
TMC_GetHeight = 2011
TMC_GetInclineSwitch = 2007
TMC_GetPrismCorr = 2023
TMC_GetRefractiveCorr = 2031
TMC_GetRefractiveMethod = 2091
TMC_GetSignal = 2022
TMC_GetSimpleCoord = 2116
TMC_GetSimpleMea = 2108
TMC_GetSlopeDistCorr = 2126
TMC_GetStation = 2009
TMC_IfDataAzeCorrError = 2114
TMC_IfDataIncCorrError = 2115
TMC_QuickDist = 2117
TMC_SetAngSwitch = 2016
TMC_SetAtmCorr = 2028
TMC_SetEdmMode = 2020
TMC_SetHandDist = 2019
TMC_SetHeight = 2012
TMC_SetInclineSwitch = 2006
TMC_SetOrientation = 2113
TMC_SetPrismCorr = 2024
TMC_SetRefractiveCorr = 2030
TMC_SetRefractiveMethod = 2090
TMC_SetStation = 2010
WIR_GetRecFormat = 8011
WIR_SetRecFormat = 8012
CRLF = "\r\n"


COMMAND_CODES = {'AUS_GetRcsSearchSwitch': 18010,
                 'AUS_GetUserAtrState': 18006,
                 'AUS_GetUserLockState': 18008,
                 'AUS_SetUserAtrState': 18005,
                 'AUS_SetUserLockState': 18007,
                 'AUS_SwitchRcsSearch': 18009,
                 'AUT_ChangeFace': 9028,
                 'AUT_ChangeFace4': 9028,
                 'AUT_FineAdjust': 9037,
                 'AUT_FineAdjust3': 9037,
                 'AUT_GetATRStatus': 9019,
                 'AUT_GetFineAdjustMode': 9030,
                 'AUT_GetLockStatus': 9021,
                 'AUT_GetSearchArea': 9042,
                 'AUT_GetUserSpiral': 9040,
                 'AUT_LockIn': 9013,
                 'AUT_MakePositioning': 9027,
                 'AUT_MakePositioning4': 9027,
                 'AUT_PS_EnableRange': 9048,
                 'AUT_PS_SearchNext': 9051,
                 'AUT_PS_SearchWindow': 9052,
                 'AUT_PS_SetRange': 9047,
                 'AUT_ReadTimeout': 9012,
                 'AUT_ReadTol': 9008,
                 'AUT_Search': 9029,
                 'AUT_Search2': 9029,
                 'AUT_SetATRStatus': 9018,
                 'AUT_SetFineAdjustMode': 9031,
                 'AUT_SetLockStatus': 9020,
                 'AUT_SetSearchArea': 9043,
                 'AUT_SetTimeout': 9011,
                 'AUT_SetTol': 9007,
                 'AUT_SetUserSpiral': 9041,
                 'BAP_GetLastDisplayedError': 17003,
                 'BAP_GetMeasPrg': 17018,
                 'BAP_GetPrismDef': 17023,
                 'BAP_GetPrismType': 17009,
                 'BAP_GetTargetType': 17022,
                 'BAP_MeasDistanceAngle': 17017,
                 'BAP_SearchTarget': 17020,
                 'BAP_SetMeasPrg': 17019,
                 'BAP_SetPrismDef': 17024,
                 'BAP_SetPrismType': 17008,
                 'BAP_SetTargetType': 17021,
                 'BMM_BeepAlarm': 11004,
                 'BMM_BeepNormal': 11003,
                 'BMM_BeepOff': 11002,
                 'BMM_BeepOn': 11001,
                 'COM_EnableSignOff': 115,
                 'COM_GetBinaryAvailable': 113,
                 'COM_GetDoublePrecision': 108,
                 'COM_GetSWVersion': 110,
                 'COM_Local': 1,
                 'COM_NullProc': 0,
                 'COM_SetBinaryAvailable': 114,
                 'COM_SetDoublePrecision': 107,
                 'COM_SetSendDelay': 109,
                 'COM_SwitchOffTPS': 112,
                 'COM_SwitchOnTPS': 111,
                 'CSV_GetDateTime': 5008,
                 'CSV_GetDeviceConfig': 5035,
                 'CSV_GetInstrumentName': 5004,
                 'CSV_GetInstrumentNo': 5003,
                 'CSV_GetIntTemp': 5011,
                 'CSV_GetSWVersion': 5034,
                 'CSV_GetSWVersion2': 5034,
                 'CSV_GetUserInstrumentName': 5006,
                 'CSV_GetVBat': 5009,
                 'CSV_GetVMem': 5010,
                 'CSV_SetDateTime': 5007,
                 'CSV_SetUserInstrumentName': 5005,
                 'CTL_GetUpCounter': 12003,
                 'EDM_GetBumerang': 1044,
                 'EDM_GetEglIntensity': 1058,
                 'EDM_GetTrkLightBrightness': 1041,
                 'EDM_GetTrkLightSwitch': 1040,
                 'EDM_Laserpointer': 1004,
                 'EDM_On': 1010,
                 'EDM_SetBumerang': 1007,
                 'EDM_SetEglIntensity': 1059,
                 'EDM_SetTrkLightBrightness': 1032,
                 'EDM_SetTrkLightSwitch': 1031,
                 'IOS_BeepOff': 20000,
                 'IOS_BeepOn': 20001,
                 'MOT_ReadLockStatus': 6021,
                 'MOT_SetVelocity': 6004,
                 'MOT_StartController': 6001,
                 'MOT_StopController': 6002,
                 'SUP_GetConfig': 14001,
                 'SUP_SetConfig': 14002,
                 'SUP_SwitchLowTempControl': 14003,
                 'TMC_DoMeasure': 2008,
                 'TMC_GetAngSwitch': 2014,
                 'TMC_GetAngle1': 2003,
                 'TMC_GetAngle5': 2107,
                 'TMC_GetAtmCorr': 2029,
                 'TMC_GetCoordinate': 2082,
                 'TMC_GetCoordinate1': 2082,
                 'TMC_GetEdmMode': 2021,
                 'TMC_GetFace': 2026,
                 'TMC_GetHeight': 2011,
                 'TMC_GetInclineSwitch': 2007,
                 'TMC_GetPrismCorr': 2023,
                 'TMC_GetRefractiveCorr': 2031,
                 'TMC_GetRefractiveMethod': 2091,
                 'TMC_GetSignal': 2022,
                 'TMC_GetSimpleCoord': 2116,
                 'TMC_GetSimpleMea': 2108,
                 'TMC_GetSlopeDistCorr': 2126,
                 'TMC_GetStation': 2009,
                 'TMC_IfDataAzeCorrError': 2114,
                 'TMC_IfDataIncCorrError': 2115,
                 'TMC_QuickDist': 2117,
                 'TMC_SetAngSwitch': 2016,
                 'TMC_SetAtmCorr': 2028,
                 'TMC_SetEdmMode': 2020,
                 'TMC_SetHandDist': 2019,
                 'TMC_SetHeight': 2012,
                 'TMC_SetInclineSwitch': 2006,
                 'TMC_SetOrientation': 2113,
                 'TMC_SetPrismCorr': 2024,
                 'TMC_SetRefractiveCorr': 2030,
                 'TMC_SetRefractiveMethod': 2090,
                 'TMC_SetStation': 2010,
                 'WIR_GetRecFormat': 8011,
                 'WIR_SetRecFormat': 8012,
                 }

# enums

# KInd of universal
class BOOLEAN_TYPE(Enum):
    FALSE = 0
    TRUE = 1

class ON_OFF_TYPE(Enum):  # on/off switch type
    OFF = 0  # Switch is off
    ON = 1  # Switch is on

# AUT
class AUT_DIRECTION(Enum):
    AUT_CLOCKWISE = 1, # direction clockwise.
    AUT_ANTICLOCKWISE = -1 # direction counter clockwise.

# Position Precision
class AUT_POSMODE(Enum):
    AUT_NORMAL = 0  # fast positioning mode
    AUT_PRECISE = 1  # exact positioning mode
    # note: can distinctly claim more time for the positioning

# Fine-adjust Position Mode
class AUT_ADJMODE(Enum):
# Possible settings of the positioning
# tolerance relating the angle- or the
# point- accuracy at the fine adjust.
    AUT_NORM_MODE = 0  # Angle tolerance
    AUT_POINT_MODE = 1  # Point tolerance
    AUT_DEFINE_MODE = 2  # System independent positioning tolerance. Set with AUT_SetTol

# Automatic Target Recognition Mode
class AUT_ATRMODE(Enum):  # Possible modes of the target recognition
    AUT_POSITION = 0  # Positioning to the hz- and v-angle
    AUT_TARGET = 1  # Positioning to a target in the environment of the hz- and v-angle.

# BAP - Measurement Modes
class BAP_MEASURE_PRG(Enum):
    BAP_NO_MEAS = 0  # no measurements, take last one
    BAP_NO_DIST = 1  # no dist. measurement, angles only
    BAP_DEF_DIST = 2  # default distance measurements, pre-defined using BAP_SetMeasPrg
    BAP_CLEAR_DIST = 5  # clear distances
    BAP_STOP_TRK = 6  # stop tracking laser

# BAP - Distance measurement programs
class BAP_USER_MEASPRG(Enum):
    BAP_SINGLE_REF_STANDARD = 0  # standard single IR distance with reflector
    BAP_SINGLE_REF_FAST = 1  # fast single IR distance with reflector
    BAP_SINGLE_REF_VISIBLE = 2  # long range distance with reflector (red laser)
    BAP_SINGLE_RLESS_VISIBLE = 3  # single RL distance, reflector free (red laser)
    BAP_CONT_REF_STANDARD = 4  # tracking IR distance with reflector
    BAP_CONT_REF_FAST = 5  # fast tracking IR distance with reflector
    BAP_CONT_RLESS_VISIBLE = 6  # fast tracking RL distance, reflector free (red)
    BAP_AVG_REF_STANDARD = 7  # Average IR distance with reflector
    BAP_AVG_REF_VISIBLE = 8  # Average long range dist. with reflector (red)
    BAP_AVG_RLESS_VISIBLE = 9  # Average RL distance, reflector free (red laser)

# BAP - Prism type definition
class BAP_PRISMTYPE(Enum):
    BAP_PRISM_ROUND = 0  # prism type: round
    BAP_PRISM_MINI = 1  # prism type: mini
    BAP_PRISM_TAPE = 2  # prism type: tape
    BAP_PRISM_360 = 3  # prism type: 360
    BAP_PRISM_USER1 = 4  # prism type: user1
    BAP_PRISM_USER2 = 5  # prism type: user2
    BAP_PRISM_USER3 = 6  # prism type: user3

# BAP - Reflector type definition
class BAP_REFLTYPE(Enum):
    BAP_REFL_UNDEF = 0  # reflector not defined
    BAP_REFL_PRISM = 1  # reflector prism
    BAP_REFL_TAPE = 2  # reflector tape

# BAP - Target type definition
class BAP_TARGET_TYPE(Enum):
    BAP_REFL_USE = 0  # with reflector
    BAP_REFL_LESS = 1  # without reflector

# COM
class COM_FORMAT(Enum):
    COM_ASCII = 0 # Force ASCII comm.
    COM_BINARY = 1 # Enable binary comm.

class COM_BAUD_RATE(Enum):
    COM_BAUD_38400 = 0
    COM_BAUD_19200 = 1  # default baud rate
    COM_BAUD_9600 = 2
    COM_BAUD_4800 = 3
    COM_BAUD_2400 = 4

class COM_TPS_STATUS(Enum):
    COM_TPS_OFF = 0  # switched off
    COM_TPS_SLEEPING = 1  # sleep mode
    COM_TPS_ONLINE = 2  # online mode
    COM_TPS_LOCAL = 3  # local mode
    COM_TPS_UNKNOWN = 4  # unknown or not initialised

# COM - Start Mode, used in COM_SwitchOnTPS(COM_TPS_STARTUP_MODE eOnMode)
class COM_TPS_STARTUP_MODE(Enum):
    COM_TPS_STARTUP_LOCAL = 0  # RPC’s enabled, local mode
    COM_TPS_STARTUP_REMOTE= 1  # RPC’s enabled, online mode

# COM - Stop Mode, used in COM_SwitchOffTPS(COM_TPS_STOP_MODE eOffMode)
class COM_TPS_STOP_MODE(Enum):
    COM_TPS_STOP_SHUT_DOWN = 0  # power down instrument
    COM_TPS_STOP_SLEEP = 1  # puts instrument into sleep state

# //? TODO: unused?
# TPS Device Precision Class
class TPS_DEVICE_CLASS(Enum):
    TPS_CLASS_1100 = 0  # TPS1000 family member, 1 mgon, 3"
    TPS_CLASS_1700 = 1  # TPS1000 family member, 0.5 mgon, 1.5"
    TPS_CLASS_1800 = 2  # TPS1000 family member, 0.3 mgon, 1"
    TPS_CLASS_5000 = 3  # TPS2000 family member
    TPS_CLASS_6000 = 4  # TPS2000 family member
    TPS_CLASS_1500 = 5  # TPS1000 family member
    TPS_CLASS_2003 = 6  # TPS2000 family member
    TPS_CLASS_5005 = 7  # TPS5000 family member
    TPS_CLASS_5100 = 8  # TPS5000 family member
    TPS_CLASS_1102 = 100  # TPS1100 family member, 2"
    TPS_CLASS_1103 = 101  # TPS1100 family member, 3"
    TPS_CLASS_1105 = 102  # TPS1100 family member, 5"
    TPS_CLASS_1101 = 103  # TPS1100 family member, 1"

# //? TODO: unused?
# TPS Device Configuration Type
class TPS_DEVICE_TYPE(Enum):
    TPS_DEVICE_T = 0x00000  # theodolite without built-in EDM
    TPS_DEVICE_TC1 = 0x00001  # tachymeter built-in
    TPS_DEVICE_TC2 = 0x00002  # tachymeter with red red laser built-in
    TPS_DEVICE_MOT = 0x00004  # motorized device
    TPS_DEVICE_ATR = 0x00008  # automatic target recognition
    TPS_DEVICE_EGL = 0x00010  # electronic guide light
    TPS_DEVICE_DB = 0x00020  # reserved
    TPS_DEVICE_DL = 0x00040  # diode laser
    TPS_DEVICE_LP = 0x00080  # laser plummet
    TPS_DEVICE_ATC = 0x00100  # autocollimination lamp
    TPS_DEVICE_LPNT= 0x00200  # Laserpointer
    TPS_DEVICE_RL_EXT = 0x00400  # Red laser with extended range
    TPS_DEVICE_SIM = 0x04000  # runs on simulation, not on hardware

# EDM - Intensity of Electronic Guidelight
class EDM_EGLINTENSITY_TYPE(Enum):
    EDM_EGLINTEN_OFF = 0,
    EDM_EGLINTEN_LOW = 1,
    EDM_EGLINTEN_MID = 2,
    EDM_EGLINTEN_HIGH = 3

# MOT - Lock Conditions
class MOT_LOCK_STATUS(Enum):
    MOT_LOCKED_OUT = 0  # locked out
    MOT_LOCKED_IN = 1  # locked in
    MOT_PREDICTION = 2  # prediction mode

# MOT - Controller Stop Mode
class MOT_STOPMODE(Enum):
    MOT_NORMAL = 0  # slow down with current acceleration
    MOT_SHUTDOWN = 1  # slow down by switch off power supply

# MOT - Controller Configuration
class MOT_MODE(Enum):
    MOT_POSIT = 0  # configured for relative postioning
    MOT_OCONST = 1  # configured for constant speed
    MOT_MANUPOS = 2  # configured for manual positioning default setting
    MOT_LOCK = 3  # configured as "Lock-In"-controller
    MOT_BREAK = 4 # configured as "Brake"-controller do not use 5 and 6
    MOT_TERM = 7  # terminates the controller task

# SUP - Automatic Shutdown Mechanism for the System
class SUP_AUTO_POWER(Enum):
    AUTO_POWER_DISABLED = 0  # deactivate the mechanism
    AUTO_POWER_SLEEP = 1  # activate sleep mechanism
    AUTO_POWER_OFF = 2  # activate shut down mechanism

# TMC - Inclination Sensor Measurement Program
class TMC_INCLINE_PRG(Enum):
    TMC_MEA_INC = 0  # Use sensor (apriori sigma)
    TMC_AUTO_INC = 1  # Automatic mode (sensor/plane)
    TMC_PLANE_INC = 2  # Use plane (apriori sigma)

# TMC - Measurement Mode
class TMC_MEASURE_PRG(Enum):
    TMC_STOP = 0  # Stop measurement program
    TMC_DEF_DIST = 1  # Default DIST-measurement program
    TMC_TRK_DIST = 2  # Distance-TRK measurement program
    TMC_CLEAR = 3  # TMC_STOP and clear data
    TMC_SIGNAL = 4  # Signal measurement (test function)
    TMC_DO_MEASURE = 6  # (Re)start measurement task
    TMC_RTRK_DIST = 8  # Distance-TRK measurement program
    TMC_RED_TRK_DIST = 10  # Red laser tracking
    TMC_FREQUENCY = 11  # Frequency measurement (test)

# EDM - Measurement Mode
class EDM_MODE(Enum):
    EDM_MODE_NOT_USED = 0  # Init value
    EDM_SINGLE_TAPE = 1  # Single measurement with tape
    EDM_SINGLE_STANDARD = 2  # Standard single measurement
    EDM_SINGLE_FAST = 3  # Fast single measurement
    EDM_SINGLE_LRANGE = 4  # Long range single measurement
    EDM_SINGLE_SRANGE = 5  # Short range single measurement
    EDM_CONT_STANDARD = 6  # Standard repeated measurement
    EDM_CONT_DYNAMIC = 7  # Dynamic repeated measurement
    EDM_CONT_REFLESS = 8  # Reflectorless repeated measurement
    EDM_CONT_FAST = 9  # Fast repeated measurement
    EDM_AVERAGE_IR = 10  # Standard average measurement
    EDM_AVERAGE_SR = 11  # Short range average measurement
    EDM_AVERAGE_LR = 12  # Long range average measurent

# Used before TPS1100 to set tracking light brightness
class EDM_TRKLIGHT_BRIGHTNESS(Enum):
    EDM_LOW_BRIGHTNESS = 0
    EDM_MEDIUM_BRIGHTNESS = 1
    EDM_HIGH_BRIGHTNESS = 2


class EDM_EGLINTENSITY_TYPE(Enum):
    EDM_EGLINTEN_OFF = 0
    EDM_EGLINTEN_LOW = 1
    EDM_EGLINTEN_MID = 2
    EDM_EGLINTEN_HIGH = 3


from ast import arg
from PyQt5.QtWidgets import QDialog
from PyQt5.QtCore import Qt
from .ui.tachy_joystick import Ui_Dialog as Ui_TachyJoystick
from tachyconnect.ts_control import Dispatcher
from tachyconnect.TachyRequest import (MOT_StartController,
                                       MOT_StopController,
                                       MOT_SetVelocity,
                                       AUT_Search,
                                       AUT_LockIn,
                                       AUT_PS_SearchNext,
                                       AUS_SetUserLockState,
                                       AUS_GetUserLockState,
                                       AUS_SetUserAtrState,
                                       EDM_SetEglIntensity,
                                       AUT_SetATRStatus,
                                       TMC_GetHeight
)
from tachyconnect.ReplyHandler import CommandChain, ChainableCommand

import tachyconnect.gc_constants as gc

class TachyJoystick(QDialog, Ui_TachyJoystick):
    MAX_SPEED = 0.42
    STEP = 0.06

    LOCKED = "🔒"
    UNLOCKED = " "
    LOCK_STATES = {'0': UNLOCKED,
                   '1': LOCKED}

    def __init__(self, dispatcher, parent=None, flags=Qt.Dialog | Qt.Tool):
        super().__init__(parent = parent, flags = flags)
        self.reject = self.accept
        self.dispatcher = dispatcher

        # THis block builds a chain of requests that are triggered after a successfull target search
        self.dispatcher.reply_handler.register_command(AUT_Search, self.searched)
        self.dispatcher.reply_handler.register_command(AUT_SetATRStatus, self.tracking)
        self.dispatcher.reply_handler.register_command(AUS_SetUserLockState, self.set_lock)

        # Other reply handling goes here
        self.dispatcher.reply_handler.register_command(AUS_GetUserLockState, self.show_lock_state)
        self.dispatcher.reply_handler.register_command(TMC_GetHeight, self.show_ref_height)
        self.dispatcher.reply_handler.register_command(AUT_PS_SearchNext, self.set_lock)
        # self.dispatcher.reply_handler.register_command(AUT_LockIn, self.get_lock_state)

        self.hzSpeed = 0
        self.vtSpeed = 0
        self.setupUi(self)
        self.connectSignalsSlots()

    ## BEGIN search reply handling
    def searched(self, *args):
        print('searched')
        self.dispatcher.send(AUT_SetATRStatus(args=[gc.ON_OFF_TYPE.ON.value], time_out=0.8).get_geocom_command())

    def tracking(self, *args):
        self.dispatcher.send(AUS_SetUserLockState(args=[gc.ON_OFF_TYPE.ON.value], time_out=0.8).get_geocom_command())

    def set_lock(self, *args):
        self.dispatcher.send(AUT_LockIn().get_geocom_command())
        self.get_lock_state()
    ## END search reply handling
    def get_ref_height(self):
        self.dispatcher.send(TMC_GetHeight(time_out=10).get_geocom_command())

    def show_ref_height(self, *args):
        self.refHeight.setText(f"{float(args[-1]):.3f}")

    def get_lock_state(self):
        self.dispatcher.send(AUS_GetUserLockState().get_geocom_command())

    def show_lock_state(self, *args):
        #print('Lock state: ' + args)
        if args[0] == '0':
            self.lockState.setText(TachyJoystick.LOCK_STATES[args[-1]])
        else:
            self.lockState.setText(TachyJoystick.LOCK_STATES[0])

    def connectSignalsSlots(self):
        self.joystickUp.clicked.connect(self.up)
        self.joystickOk.clicked.connect(self.accept)
        self.joystickDown.clicked.connect(self.down)
        self.joystickStop.clicked.connect(self.stop)
        self.joystickLeft.clicked.connect(self.left)
        self.joystickRight.clicked.connect(self.right)
        self.joystickLock.clicked.connect(self.lock)
        self.pwrSearch.clicked.connect(self.start_pwr_search)

    def lock(self):
        self.get_lock_state()
        self.dispatcher.send(AUT_Search(time_out=30,args=[.1, .3]).get_geocom_command())

    def lights_off(self):
        self.dispatcher.send(EDM_SetEglIntensity(args=['0']).get_geocom_command())

    def show(self):
        super().show()
        self.dispatcher.send(MOT_StartController(args=['1']).get_geocom_command())
        self.dispatcher.send(EDM_SetEglIntensity(args=['3']).get_geocom_command())
        self.get_lock_state()
        self.get_ref_height()


    def accept(self):
        self.stop()
        self.dispatcher.send(MOT_StopController().get_geocom_command())
        self.lights_off()
        super().accept()

    def set_velocity(self):
        self.dispatcher.send(MOT_SetVelocity(args=[self.hzSpeed, self.vtSpeed]).get_geocom_command())

    def up(self):
        if abs(self.vtSpeed) < TachyJoystick.MAX_SPEED:
            self.vtSpeed -= TachyJoystick.STEP
            self.set_velocity()

    def down(self):
        if abs(self.vtSpeed) < TachyJoystick.MAX_SPEED:
            self.vtSpeed += TachyJoystick.STEP
            self.set_velocity()

    def left(self):
        if abs(self.hzSpeed) < TachyJoystick.MAX_SPEED:
            self.hzSpeed += TachyJoystick.STEP
            self.set_velocity()

    def right(self):
        if abs(self.hzSpeed) < TachyJoystick.MAX_SPEED:
            self.hzSpeed -= TachyJoystick.STEP
            self.set_velocity()

    def stop(self):
        self.hzSpeed = self.vtSpeed = 0
        self.set_velocity()

    def start_pwr_search(self):
        self.dispatcher.send(AUT_PS_SearchNext(time_out=100,args=[d.value for d in AUT_PS_SearchNext.defaults]).get_geocom_command())


from datetime import datetime as dt
from . import gc_constants as gc
from .ts_control import GeoCOMCommand, GSICommand, CommunicationConstants
from PyQt5.QtCore import QObject

class TachyRequest(QObject):
    gsi_command = ""
    gc_command = ""
    unpacking_keys = {}
    defaults = []
    description = ""

    def __init__(self, time_out = 2, args = []):
        super().__init__()
        self.gc_command = str(gc.COMMAND_CODES.get(self.get_class_name()))
        self.time_out = time_out
        self.args = args

    def get_class_name(self):
        return self.__class__.__name__

    def __str__(self):
        return f"{self.get_class_name()}: {', '.join(self.args) if len(self.args) else 'No args'}, {self.time_out} seconds"

    def get_gsi_command(self):
        if self.gsi_command == "":
            raise NotImplementedError(f'No GSI command for {self.get_class_name()}')
        return GSICommand(self.gsi_command, self.get_class_name(), self.time_out, *self.args)

    def get_geocom_command(self):
        return GeoCOMCommand(self.gc_command, self.get_class_name(), self.time_out, *self.args)

    @classmethod
    def get_defaults(cls):
        return cls.defaults

    @classmethod
    def get_helptext(cls, i):
        if i < len(cls.helptexts):
            return cls.helptexts[i]
        else:
            return ''


class COM_NullProc(TachyRequest):
    description = """Check Communication
This function does not provide any functionality except of checking if the
communication is up and running."""


class COM_Local(TachyRequest):
    description = """Switch TPS1100 into Local Mode
Leaves on-line mode and switches TPS1100 into local mode. If in local
mode, no communication will take place. Any attempt of sending data will
be ignored. Changing local into online mode can be done manually only."""


class COM_SetDoublePrecision(TachyRequest):
    description = """Set Double Precision Setting
This function sets the precision - number of digits to the right of the
decimal - when double floating-point values are transmitted. The TPS’
system software always calculates with highest possible precision. The
default precision is fifteen digits. However, if this precision is not needed
then transmission of double data (ASCII transmission) can be speeded up
by choosing a lower precision. Especially when many double values are
transmitted this may enhance the operational speed. The usage of this
function is only meaningful if the communication is set to ASCII
transmission mode. In the case of an ASCII request, the precision of the
server side will be set. Notice that trailing Zeros will not be sent by the
server and values may be rounded. E.g. if precision is set to 3 and the exact
value is 1.99975 the resulting value will be 2.0
Note: With this function one can decrease the accuracy of the delivered
values."""
    defaults = ["15"]
    helptexts = ["Number of digits right to the comma"]


class COM_GetDoublePrecision(TachyRequest):
    description = """Get Double Precision Setting
This function returns the precision - number of digits to the right of the
decimal point - when double floating-point values are transmitted. The
usage of this function is only meaningful if the communication is set to
ASCII transmission mode. Precision is equal in both transmission
directions. In the case of an ASCII request, the precision of the server side
will be returned."""


class COM_SetSendDelay(TachyRequest):
    description = """Set Reply Delay
The GeoCOM implementation of the server has been optimised for speed.
If the server reacts to fast, then it may happen, that the client is not able to
receive the reply (complete and) correctly. This RPC inserts a delay before
the server responds to a request. This might be of interest especially for
radio data links. Reset to no delay can be done with nSendDelay = 0."""
    defaults = ["0"]
    helptexts = ["Time of transmission delay in milliseconds"]


class COM_GetBinaryAvailable(TachyRequest):
    description = """Get Binary Attribute of Server
This function gets the ability information about the server to handle binary
communication. Since TPS1100 Release 2.00 the client may make requests in
binary format which speeds up the communication by about 40-50%."""


class COM_SetBinaryAvailable(TachyRequest):
    description = """Set Binary Attribute of Server
This function sets the ability of the server to handle binary communication.
With this function, one can force to communicate in ASCII only. During
initialisation, the client checks if binary communication is enabled /
possible or not which depends on this flag. Binary data format is not
supported yet in GeoCOM Versions below 2.0."""
    defaults = [gc.BOOLEAN_TYPE.TRUE]
    helptexts = ["""TRUE: enable binary operation.
FALSE: enable ASCII operation only."""]


class COM_GetSWVersion(TachyRequest):
    gsi_command = "GET/I/WI593;"
    description = """Retrieve Server Release Information
This function retrieves the current GeoCOM release (release, version and
subversion) of the server."""


class COM_SwitchOnTPS(TachyRequest):
    gsi_command = "a"
    description = """Switch on TPS instrument
This function switches on the TPS1100 instrument and put it into remote
mode. It can also be used to switch from sleep into remote mode."""
    defaults = [gc.COM_TPS_STARTUP_MODE.COM_TPS_STARTUP_REMOTE]
    helptexts = """Run mode - use COM_TPS_STARTUP_REMOTE only!
COM_TPS_STARTUP_LOCAL will yield to erroneous behaviour."""


class COM_SwitchOffTPS(TachyRequest):
    gsi_command = "b"
    description = """Switch off TPS1100 or Set Sleep Mode
This function switches off the TPS1100 instrument or put it into sleep
mode."""
    defaults = [gc.COM_TPS_STOP_MODE.COM_TPS_STOP_SLEEP]
    helptexts = ["Stop mode"]


class COM_EnableSignOff(TachyRequest):
    description = """Enable Remote Mode Logging
This function enables logging if the Remote mode changes. See also
section 3.5 TPS1100 Instrument Modes of Operation for further
explanations."""
    defaults = [gc.BOOLEAN_TYPE.FALSE]
    helptexts = ["TRUE: enable mode logging"]


class EDM_Laserpointer(TachyRequest):
    description = """Switch on/off laserpointer
Laserpointer is only available in theodolites which supports distance
measurement without reflector."""
    defaults = [gc.ON_OFF_TYPE.ON]
    helptexts = ["""
ON - switch Laserpointer on
OFF - switch Laserpointer off"""]


class EDM_SetBumerang(TachyRequest):
    description = """Switch boomerang filter on/off
The boomerang filter is not available for some add-on EDM’s.
Deleted in TPS1100"""
    defaults = [gc.ON_OFF_TYPE.OFF]
    helptexts = ["""ON - switch boomerang filter on
OFF - switch boomerang filter off"""]


class EDM_On(TachyRequest):
    description = """Switch on/off EDM
Normal distance measurement switches on or off the EDM automatically. If
there is no supply voltage for the EDM, EDM_On also connects it. Normally
the supply voltage is switched off together with the EDM, unless the
Tracklight or the diode laser is switched on too, or the permanent power-on-
mode is set. These functions overlie the functionality of EDM_On and EDM
cannot be switched off before the Tracklight or the diode laser stops
working. However, after 10 min the EDM will be switched off.
Deleted in TPS1100"""
    defaults = [gc.ON_OFF_TYPE.ON]
    helptexts = ["""ON - switch EDM on
OFF - switch EDM off"""]


class EDM_SetTrkLightSwitch(TachyRequest):
    description = """Switch on/off tracklight
The Tracklight must be available for EDM.
Replaced by: EDM_SetEGLIntensity in TPS1100"""
    defaults = [gc.ON_OFF_TYPE.ON]
    helptexts = ["""ON - switch on Tracklight
OFF - switch off Tracklight"""]


class EDM_SetTrkLightBrightness(TachyRequest):
    description = """Change intensity of tracklight
The Tracklight must be available for EDM.
Replaced by: EDM_SetEGLIntensity in TPS1100"""
    defaults = [gc.EDM_TRKLIGHT_BRIGHTNESS.EDM_HIGH_BRIGHTNESS]
    helptexts = [""""""]


class EDM_GetTrkLightSwitch(TachyRequest):
    description = """Get status of tracklight switch
The Tracklight must be available for EDM.
Replaced by: EDM_GetEGLIntensity in TPS1100"""


class EDM_GetTrkLightBrightness(TachyRequest):
    description = """Get value of intensity of tracklight
The Tracklight must be available for EDM.
Replaced by: EDM_GetEGLIntensity"""


class EDM_GetBumerang(TachyRequest):
    description = """Get status of boomerang filter
Call this function to retrieve the status of the "boomerang filter" (i.e. ON,
OFF). If the distance is within 60-100 meters and the boomerang filter is
turned off, wrong measurement results are possible. With the boomerang
filter turned on, the measurement accuracy can be increased until up to max.
2.5 cm.
The boomerang filter is not available for some add-on EDM’s.
Deleted in TPS1100"""


class EDM_GetEglIntensity(TachyRequest):
    description = """Get value of intensity of guide light
The Electronic Guide Light must be implemented in the theodolite."""


class EDM_SetEglIntensity(TachyRequest):
    description = """Change intensity of guide light
The Electronic Guide Light must be implemented in the theodolite.
New since TPS1100"""
    defaults=[gc.EDM_EGLINTENSITY_TYPE.EDM_EGLINTEN_HIGH]
    helptexts=[""""""]


class TMC_GetAngle1(TachyRequest):
    description = """Returns complete angle measurement
This function carries out an angle measurement and, in dependence of
configuration, inclination measurement and returns the results. As shown
the result is very comprehensive. For simple angle measurements use
TMC_GetAngle5 or TMC_GetSimpleMea instead."""
    defaults = [gc.TMC_INCLINE_PRG.TMC_AUTO_INC]
    helptexts = ["Inclination sensor measurement mode."]


class TMC_SetInclineSwitch(TachyRequest):
    description = """Switch dual axis compensator on or off
This function switches the dual axis compensator on or off."""
    defaults = [gc.ON_OFF_TYPE.ON]
    helptexts = ["Dual axis compensator's state."]


class TMC_GetInclineSwitch(TachyRequest):
    description = """Get the dual axis compensator's state
This function returns the current dual axis compensator's state."""


class TMC_DoMeasure(TachyRequest):
    description = """Carries out a distance measurement
This function carries out a distance measurement in a variety of TMC
measurement modes like single distance, rapid tracking,... . Please note that
this command does not output any values (distances). In order to get the
values you have to use other measurement functions such as
TMC_GetCoordinate, TMC_GetSimpleMea or TMC_GetAngle.
The value of the distance measured is kept in the instrument up to the next
TMC_DoMeasure command where a new distance is requested or the
distance is clear by the measurement program TMC_CLEAR.
Note: If you perform a distance measurement with the measure program
TMC_DEF_DIST, the distance sensor will be work with the set EDM
mode, see TMC_SetEdmMode."""
    defaults = [gc.TMC_MEASURE_PRG.TMC_DO_MEASURE, gc.TMC_INCLINE_PRG.TMC_AUTO_INC]
    helptexts = ["TMC measurement mode.", "Inclination sensor measurement mode."]


class TMC_GetStation(TachyRequest):
    description = """Get the coordinates of the instrument station
This function is used to get the co-ordinates of the instrument station."""


class TMC_SetStation(TachyRequest):
    description = """Set the coordinates of the instrument station
This function is used to set the co-ordinates of the instrument station."""
    defaults = ["0", "0", "0", "0"]
    helptexts = ["E0[double] - Station easting coordinate",
               "N0[double] - Station northing coordinate",
               "H0[double] - Station height coordinate",
               "Hi[double] - Instrument height"]


class TMC_GetHeight(TachyRequest):
    gsi_command = "GET/I/WI88"
    description = """"Returns the current reflector height
This function returns the current reflector height."""



class TMC_SetHeight(TachyRequest):
    description = """Sets new reflector height
This function sets a new reflector height."""
    defaults = ["0"]
    helptexts = ["New reflector height"]


class TMC_GetAngSwitch(TachyRequest):
    description = """Get angular correction's states
This function returns the angular correction's state."""


# ASCII Request takes no parameters
class TMC_SetAngSwitch(TachyRequest):
    description = """Enable/disable angle corrections
With this function you can enable/disable follow angle measurement correction.
incline: The incline will be considered in the angle
measurement if enabled.
stand axis: The stand axis will be considered in the angle
measurement if enabled.
collimation: The collimation will be considered in the angle
measurement if enabled
tilt axis: The tilt axis will be considered in the angle
measurement if enabled.
Note: You can set the various corrections only, if no distance is existing! (Use the
function TMC_DoMeasure(TMC_CLEAR,..) in order to clear the distance)"""


class TMC_SetHandDist(TachyRequest):
    description = """Input slope distance and height offset
This function is used to input manually measured slope distance and height
offset for a following measurement. Additionally an inclination
measurement and an angle measurement are carried out to determine the
co-ordinates of target. The V-angle is corrected to π/2 or 3⋅π/2 in
dependence of the instrument’s face because of the manual input.
After the function call the previous measured distance is cleared."""
    defaults = ["10", "1", gc.TMC_INCLINE_PRG.TMC_AUTO_INC]
    helptexts = ["Slope distance", "Height offset" , "Inclination sensor measurement mode."]


class TMC_SetEdmMode(TachyRequest):
    description = """Set EDM measurement modes
This function set the current measurement modes new. The measure
function TMC_DoMeasure(TMC_DEF_DIST) will work with this
configuration."""
    defaults = [gc.EDM_MODE.EDM_SINGLE_STANDARD]
    helptexts = ["EDM measurement mode."]


class TMC_GetEdmMode(TachyRequest):
    description = """Get the EDM measurement mode
This function returns the EDM measurement mode."""


class TMC_GetSignal(TachyRequest):
    description = """Get information about EDM’s signal amplitude
This function returns information about the amplitude of the EDM signal.
The function only can perform measuring if the signal measurement
program is activated. Start the signal measurement program with
TMC_DoMeasure where Command = TMC_SIGNAL. After the
measurement the EDM must be switch off (use TMC_DoMeasure where
Command = TMC_CLEAR"""


class TMC_GetPrismCorr(TachyRequest):
    description = """Get the prism constant
This function is used to get the prism constant."""


class TMC_SetPrismCorr(TachyRequest):
    description = """Set the prism constant
This function is used to set the prism constant.
The high-level function BAP_SetPrismType overwrites this setting."""
    defaults = ["0.1"]
    helptexts = ["Prism constant [mm]"]


class TMC_GetFace(TachyRequest):
    description = """Get face information of current telescope position
This function returns the face information of the current telescope position.
The face information is only valid, if the instrument is in an active
measurement state (that means a measurement function was called before
the TMC_GetFace call, see example). Note that the instrument
automatically turns into an inactive measurement state after a predefined
timeout."""

# todo: Try GetAtmCorr for default values
class TMC_SetAtmCorr(TachyRequest):
    description = """Set atmospheric correction parameters
This function is used to set the parameters for the atmospheric correction."""
    defaults = ["", "", "", ""]
    helptexts = ["Wave length of the EDM transmitter",
                "Atmospheric pressure",
                "Dry temperature",
                "Wet temperature"]


class TMC_GetAtmCorr(TachyRequest):
    description = """Get atmospheric correction parameters
This function is used to get the parameters for the atmospheric correction."""


# todo: Try GetRefractiveCorr for default values
class TMC_SetRefractiveCorr(TachyRequest):
    description = """Set the refraction factor
This function is used to set the refraction distortion factor for correction of
measured height difference."""
    defaults = [gc.ON_OFF_TYPE.ON, "", ""]
    helptexts = ["Refraction correction On/Off", "Radius of the earth", "Refractive coefficient"]


class TMC_GetRefractiveCorr(TachyRequest):
    gsi_command = "GET/I/WI538"
    description = """Get the refraction factor
This function is used to get the refraction distortion factor for correction of
measured height difference."""


class TMC_GetCoordinate(TachyRequest):
    description = """Gets the coordinates of a measured point
This function issues an angle measurement and, in dependence of the
selected Mode, an inclination measurement and calculates the co-ordinates
of the measured point with an already measured distance. The WaitTime is
a delay to wait for the distance measurement to finish. Single and tracking
measurements are supported. Information about a missing distance
measurement and other information about the quality of the result is
returned in the return-code."""
    defaults=["1000", gc.TMC_INCLINE_PRG.TMC_AUTO_INC]
    helptexts=["""The delay to wait for the distance
measurement to finish [ms].""",
"Inclination sensor measurement mode"]


class TMC_GetCoordinate1(TMC_GetCoordinate):
    pass


class TMC_SetRefractiveMethod(TachyRequest):
    description = """Set the refraction model
This function is used to set the refraction model."""
    defaults=["2"]
    helptexts=["""Refraction data:
Method = 1 means method 1 (for
Australia)
Method = 2 means method 2 (for the rest
of the world)"""]


class TMC_GetRefractiveMethod(TachyRequest):
    description = """Get the refraction model
This function is used to get the current refraction model."""


class TMC_GetAngle5(TachyRequest):
    description = """Returns simple angle measurement
This function carries out an angle measurement and returns the results. In
contrast to the function TMC_GetAngle1 this function returns only the
values of the angle. For simple angle measurements use or
TMC_GetSimpleMea instead.
Information about measurement is returned in the return code."""


class TMC_GetSimpleMea(TachyRequest):
    description = """Returns angle and distance measurement
This function returns the angles and distance measurement data. The
distance measurement will be set invalid afterwards. It is important to note
that this command does not issue a new distance measurement.
If a distance measurement is valid the function ignores WaitTime and
returns the results.
If no valid distance measurement is available and the distance measurement
unit is not activated (by TMC_DoMeasure before the TMC_GetSimpleMea
call) the WaitTime is also ignored and the angle measurement result is
returned. So this function can be used instead of TMC_GetAngle5.
Information about distance measurement is returned in the return- code."""
    defaults=["3000", gc.TMC_INCLINE_PRG.TMC_AUTO_INC]
    helptexts=["""The delay to wait for the distance
measurement to finish [ms].""", """Inclination sensor measurement mode."""]


class TMC_SetOrientation(TachyRequest):
    description = """Orients the theodolite in Hz direction
This function is used to orientates the instrument in Hz direction. It is a
combination of an angle measurement to get the Hz offset and afterwards
setting the angle Hz offset in order to orientates onto a target. Before the
new orientation can be set an existing distance must be cleared (use
TMC_DoMeasure with the command = TMC_CLEAR)."""
    defaults=["0.0"]
    helptexts=["Hz Orientation [rad]"]


class TMC_IfDataAzeCorrError(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class TMC_IfDataIncCorrError(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class TMC_GetSimpleCoord(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class TMC_QuickDist(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class TMC_GetSlopeDistCorr(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class CSV_GetInstrumentNo(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class CSV_GetInstrumentName(TachyRequest):
    gsi_command = "GET/I/WI13"


class CSV_SetUserInstrumentName(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class CSV_GetUserInstrumentName(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class CSV_SetDateTime(TachyRequest):
    helptexts = ["Year", "Month", "Day", "Hour", "Minute", "Second"]

    @classmethod
    def get_defaults(cls):
        return dt.now().strftime("%Y,%m,%d,%H,%M,%S").split(",")

class CSV_GetDateTime(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class CSV_GetVBat(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class CSV_GetVMem(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class CSV_GetIntTemp(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class CSV_GetSWVersion(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class CSV_GetSWVersion2(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class CSV_GetDeviceConfig(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class MOT_StartController(TachyRequest):
    description = """Start motor controller
This command is used to enable remote or user interaction to the motor
controller."""
    defaults = [gc.MOT_MODE.MOT_MANUPOS]
    helptexts = ["""Controller mode. If used together with
MOT_SetVelocity the control mode has
to be MOT_OCONST"""]


class MOT_StopController(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class MOT_SetVelocity(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class MOT_ReadLockStatus(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class WIR_GetRecFormat(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class WIR_SetRecFormat(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class AUT_SetTol(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class AUT_ReadTol(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class AUT_SetTimeout(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class AUT_ReadTimeout(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class AUT_LockIn(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class AUT_SetATRStatus(TachyRequest):
    description = ""
    defaults = [gc.ON_OFF_TYPE.ON]
    helptexts = ['On or off']


class AUT_GetATRStatus(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class AUT_SetLockStatus(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class AUT_GetLockStatus(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class AUT_MakePositioning(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class AUT_MakePositioning4(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class AUT_ChangeFace(TachyRequest):
    description = """"""
    defaults = [gc.AUT_POSMODE.AUT_NORMAL,
                gc.AUT_ATRMODE.AUT_TARGET,
                gc.BOOLEAN_TYPE.FALSE]
    helptexts = ["""\
Position mode:
AUT_NORMAL: uses the current value of the compensator.
For values >25GON positioning might tend to inexact.
AUT_PRECISE: tries to measure exact inclination of target.
Tends to long position time
(check AUT_TIMEOUT and/or COM-time out if necessary).""",
"""Mode of ATR:
AUT_POSITION: conventional position to other face.
AUT_TARGET: tries to position onto a target in the destination area.
This set is only possible if ATR exists and is activated.""",
"""It’s reserved for future use, set bDummy always to FALSE"""]


class AUT_ChangeFace4(AUT_ChangeFace):
    description = """"""
    defaults = [""""""]
    helptexts = [""""""]


class AUT_Search(TachyRequest):
    description = """"""
    defaults = ["0.1", "0.3"]
    helptexts = ["Horizontal search angle in rad.",
                "Vertical search angle in rad."]


class AUT_Search2(TachyRequest):
    description = """"""
    defaults = [""""""]
    helptexts = [""""""]


class AUT_GetFineAdjustMode(TachyRequest):
    description = """"""
    defaults = [""""""]
    helptexts = [""""""]


class AUT_SetFineAdjustMode(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class AUT_FineAdjust(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class AUT_FineAdjust3(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class AUT_GetUserSpiral(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class AUT_SetUserSpiral(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class AUT_GetSearchArea(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class AUT_SetSearchArea(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class AUT_PS_SetRange(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class AUT_PS_EnableRange(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class AUT_PS_SearchNext(TachyRequest):
    description = """Searching for the next target
This command executes the 360º default PowerSearch and searches for the next target. A previously defined
PowerSearch window (AUT_SetSearchArea) is not taken into account. Use AUT_PS_SearchWindow to do so.
TPS1200+"""
    defaults = [gc.AUT_DIRECTION.AUT_ANTICLOCKWISE, gc.BOOLEAN_TYPE.TRUE]
    helptexts = ["Defines the searching direction (CLKW=1 or ACLKW=-1)", """TRUE: Searching starts -10 gon to the given direction
lDirection. This setting finds targets left of the telescope
direction faster"""]


class AUT_PS_SearchWindow(TachyRequest):
    description = """"""
    defaults = [""""""]
    helptexts = [""""""]


class BMM_BeepOn(TachyRequest):
    description = """"""
    defaults = [""""""]
    helptexts = [""""""]


class BMM_BeepOff(TachyRequest):
    description = """"""
    defaults = [""""""]
    helptexts = [""""""]


class BMM_BeepNormal(TachyRequest):
    description = """"""
    defaults = [""""""]
    helptexts = [""""""]


class BMM_BeepAlarm(TachyRequest):
    description = """"""
    defaults = [""""""]
    helptexts = [""""""]


class CTL_GetUpCounter(TachyRequest):
    description = """"""
    defaults = [""""""]
    helptexts = [""""""]


class SUP_GetConfig(TachyRequest):
    description = """"""
    defaults = [""""""]
    helptexts = [""""""]


class SUP_SetConfig(TachyRequest):
    description = """"""
    defaults = [""""""]
    helptexts = [""""""]


class SUP_SwitchLowTempControl(TachyRequest):
    description = """"""
    defaults = [""""""]
    helptexts = [""""""]


class BAP_GetLastDisplayedError(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class BAP_SetPrismType(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class BAP_GetPrismType(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class BAP_MeasDistanceAngle(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class BAP_GetMeasPrg(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class BAP_SetMeasPrg(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class BAP_SearchTarget(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class BAP_SetTargetType(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class BAP_GetTargetType(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class BAP_GetPrismDef(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class BAP_SetPrismDef(TachyRequest):
    description = """"""
    defaults=[""""""]
    helptexts=[""""""]


class AUS_GetUserAtrState(TachyRequest):
    description = """Get the status of the ATR mode
Get the current status of the ATR mode on TCA instruments. This
command does not indicate whether the ATR has currently acquired a
prism. It replaces the function AUT_GetAtrStatus."""


class AUS_SetUserAtrState(TachyRequest):
    description = """Set the status of the ATR mode
Activate respectively deactivate the ATR mode.
Activate ATR mode:
The ATR mode is activated and the LOCK mode (if sets) will be reset
automatically also."""
    defaults = [gc.ON_OFF_TYPE.OFF]
    helptexts = ["State of the ATR mode"]


class AUS_GetUserLockState(TachyRequest):
    description = """Get the status of the lock switch
This command gets the current LOCK switch. This command is valid for
TCA instruments only and does not indicate whether the ATR has a prism
in lock or not.
With the function MOT_ReadLockStatus you can find out whether a
target is locked or not.
This command is valid for TCA instruments only. It replaces the function
AUT_GetLockStatu"""


class AUS_SetUserLockState(TachyRequest):
    description = """Set the lock status.
Status ON:
The target tracking functionality is available but not activated. In order to
activate target tracking, see the function AUT_LockIn. The ATR mode will
be set automatically.
Status OFF:
A running target tracking will be aborted and the manual driving wheel is
activated. The ATR mode will be not reset automatically respectively keep
unchanged.
This command is valid for TCA instruments only. It replaces the function
AUT_SetLockStatus."""
    defaults = [gc.ON_OFF_TYPE.OFF]
    helptexts = ["State of the ATR lock switch"]


class AUS_SwitchRcsSearch(TachyRequest):
    description = """Set the RCS searching mode switch.
If the RCS style searching is enabled, then the extended for
BAP_SearchTarget or after a loss of lock is activated.
This command is valid for TCA instruments only."""
    defaults = [gc.ON_OFF_TYPE.OFF]
    helptexts = ["Get RCS-Searching mode switch"]


class AUS_GetRcsSearchSwitch(TachyRequest):
    description = """Get RCS-Searching mode switch
This command gets the current RCS-Searching mode switch.
If RCS style searching is enabled, then the extended searching for
BAP_SearchTarget or after a loss of lock is activated.
This command is valid for TCA instruments only."""


class IOS_BeepOn(TachyRequest):
    description = """Start a beep-signal
This function switches on the beep-signal with the intensity nIntens. If a
continuous signal is active, it will be stopped first. Turn off the beeping
device with IOS_BeepOff"""
    defaults = ["100"]
    helptexts = ["""Intensity of the beep-signal (volume)
expressed as a percentage.
Default value is 100 %"""]


class IOS_BeepOff(TachyRequest):
    description = """Stop active beep-signal
This function switches off the beep-signal."""


ALL_COMMANDS = [
    COM_NullProc,
    COM_Local,
    COM_SetDoublePrecision,
    COM_GetDoublePrecision,
    COM_SetSendDelay,
    COM_GetSWVersion,
    COM_SwitchOnTPS,
    COM_SwitchOffTPS,
    COM_GetBinaryAvailable,
    COM_SetBinaryAvailable,
    COM_EnableSignOff,
    EDM_Laserpointer,
    EDM_SetBumerang,
    EDM_On,
    EDM_SetTrkLightSwitch,
    EDM_SetTrkLightBrightness,
    EDM_GetTrkLightSwitch,
    EDM_GetTrkLightBrightness,
    EDM_GetBumerang,
    EDM_GetEglIntensity,
    EDM_SetEglIntensity,
    TMC_GetAngle1,
    TMC_SetInclineSwitch,
    TMC_GetInclineSwitch,
    TMC_DoMeasure,
    TMC_GetStation,
    TMC_SetStation,
    TMC_GetHeight,
    TMC_SetHeight,
    TMC_GetAngSwitch,
    TMC_SetAngSwitch,
    TMC_SetHandDist,
    TMC_SetEdmMode,
    TMC_GetEdmMode,
    TMC_GetSignal,
    TMC_GetPrismCorr,
    TMC_SetPrismCorr,
    TMC_GetFace,
    TMC_SetAtmCorr,
    TMC_GetAtmCorr,
    TMC_SetRefractiveCorr,
    TMC_GetRefractiveCorr,
    TMC_GetCoordinate,
    TMC_GetCoordinate1,
    TMC_SetRefractiveMethod,
    TMC_GetRefractiveMethod,
    TMC_GetAngle5,
    TMC_GetSimpleMea,
    TMC_SetOrientation,
    TMC_IfDataAzeCorrError,
    TMC_IfDataIncCorrError,
    TMC_GetSimpleCoord,
    TMC_QuickDist,
    TMC_GetSlopeDistCorr,
    CSV_GetInstrumentNo,
    CSV_GetInstrumentName,
    CSV_SetUserInstrumentName,
    CSV_GetUserInstrumentName,
    CSV_SetDateTime,
    CSV_GetDateTime,
    CSV_GetVBat,
    CSV_GetVMem,
    CSV_GetIntTemp,
    CSV_GetSWVersion,
    CSV_GetSWVersion2,
    CSV_GetDeviceConfig,
    MOT_StartController,
    MOT_StopController,
    MOT_SetVelocity,
    MOT_ReadLockStatus,
    WIR_GetRecFormat,
    WIR_SetRecFormat,
    AUT_SetTol,
    AUT_ReadTol,
    AUT_SetTimeout,
    AUT_ReadTimeout,
    AUT_LockIn,
    AUT_SetATRStatus,
    AUT_GetATRStatus,
    AUT_SetLockStatus,
    AUT_GetLockStatus,
    AUT_MakePositioning,
    AUT_MakePositioning4,
    AUT_ChangeFace,
    AUT_ChangeFace4,
    AUT_Search,
    AUT_Search2,
    AUT_GetFineAdjustMode,
    AUT_SetFineAdjustMode,
    AUT_FineAdjust,
    AUT_FineAdjust3,
    AUT_GetUserSpiral,
    AUT_SetUserSpiral,
    AUT_GetSearchArea,
    AUT_SetSearchArea,
    AUT_PS_SetRange,
    AUT_PS_EnableRange,
    AUT_PS_SearchNext,
    AUT_PS_SearchWindow,
    BMM_BeepOn,
    BMM_BeepOff,
    BMM_BeepNormal,
    BMM_BeepAlarm,
    CTL_GetUpCounter,
    SUP_GetConfig,
    SUP_SetConfig,
    SUP_SwitchLowTempControl,
    BAP_GetLastDisplayedError,
    BAP_SetPrismType,
    BAP_GetPrismType,
    BAP_MeasDistanceAngle,
    BAP_GetMeasPrg,
    BAP_SetMeasPrg,
    BAP_SearchTarget,
    BAP_SetTargetType,
    BAP_GetTargetType,
    BAP_GetPrismDef,
    BAP_SetPrismDef,
    AUS_SetUserAtrState,
    AUS_GetUserAtrState,
    AUS_SetUserLockState,
    AUS_GetUserLockState,
    AUS_SwitchRcsSearch,
    AUS_GetRcsSearchSwitch,
    IOS_BeepOff,
    IOS_BeepOn]

if __name__=="__main__":
    set_station = TMC_SetStation()
    print(f"Code for set station: {set_station.gc_command}")

